import argparse
import datetime
import os
import shutil

import astropy.units as u
from astropy.io import fits
from dateutil.parser import parse
from sunpy.net import Fido
from sunpy.net import attrs as a
from tqdm import tqdm

parser = argparse.ArgumentParser(description='Download aligned STEREO A+B data')
parser.add_argument('--download_dir', type=str, help='path to the download directory.')

args = parser.parse_args()
download_dir = args.download_dir

wavelengths = [171, 195, 284, 304, ]
t_start = datetime.datetime(2016, 4, 1)
t_end = datetime.datetime.now()
td = datetime.timedelta(days=30)

times = [t_start + i * td for i in range((t_end - t_start) // td)]
[os.makedirs(os.path.join(download_dir, str(wl)), exist_ok=True) for wl in wavelengths]

def round_hour(t):
    # Rounds to nearest hour by adding a timedelta hour if minute >= 30
    return (t.replace(second=0, microsecond=0, minute=0, hour=t.hour)
            + datetime.timedelta(hours=t.minute // 30))

for time in tqdm(times):
    queries = []
    for wl in wavelengths:
        stereo_a = Fido.search(a.Instrument("EUVI"), a.Source('STEREO_A'), a.Time(time, time + td), a.Sample(1 * u.day),
                               a.Wavelength(wl * u.AA))
        if time.year > 2014: # loss of STEREO B
            queries += [stereo_a]
            continue
        stereo_b = Fido.search(a.Instrument("EUVI"), a.Source('STEREO_B'), a.Time(time, time + td), a.Sample(1 * u.day),
                               a.Wavelength(wl * u.AA))
        queries += [stereo_a, stereo_b]

    files = Fido.fetch(*queries, path=download_dir, progress=False)
    for file in files:
        if not os.path.exists(file):
            continue
        header = fits.getheader(file)
        if header['NAXIS1'] < 2048 or header['NAXIS2'] < 2048 or header['NMISSING'] != 0:
            os.remove(file)
            print('invalid file:', file, header['NAXIS1'], header['NAXIS2'], header['NMISSING'])
            continue
        wl = int(header['wavelnth'])
        s_id = 'A' if header['obsrvtry'] == 'STEREO_A' else 'B'
        if 'date-obs' not in header:
            header['date-obs'] = header['date_obs']
        d = round_hour(parse(header['date-obs']))
        target_path = os.path.join(download_dir, str(wl),
                                   '%s_%s.fits' % (d.isoformat('T'), s_id))
        print(file, target_path)
        shutil.move(file, target_path)



from itipy.data.dataset import HMIContinuumDataset, StorageDataset, HinodeDataset, SDODataset, SOHODataset, STEREODataset, \
    get_intersecting_files

if __name__ == '__main__':
    months = [2, 3, 4, 5, 6, 7, 8, 9, 11, 12]
    base_path = '/gpfs/gpfs0/robert.jarolim/data'
    ###########################################################
    # HMI
    ###########################################################
    hmi_dataset = HMIContinuumDataset('%s/iti/hmi_continuum' % base_path, (512, 512))
    hmi_dataset = StorageDataset(hmi_dataset, '%s/converted/hmi_continuum' % base_path, ext_editors=[])
    hmi_dataset.convert(12)
    ###########################################################
    # Hinode
    ###########################################################
    hinode_dataset = HinodeDataset("%s/iti/hinode_iti2022_prep" % base_path)
    hinode_dataset = StorageDataset(hinode_dataset, '%s/converted/hinode_continuum' % base_path,
                                    ext_editors=[])
    hinode_dataset.convert(12)
    ###########################################################
    # SDO
    ###########################################################
    sdo_files = get_intersecting_files("%s/iti/sdo" % base_path, ['171', '193', '211', '304', '6173'],
                                       ext='.fits', months=months)

    sdo_dataset = SDODataset(sdo_files, resolution=2048, patch_shape=(1024, 1024))
    sdo_dataset = StorageDataset(sdo_dataset, '%s/converted/sdo_2048' % base_path, ext_editors=[])
    sdo_dataset.convert(12)
    ###########################################################
    sdo_dataset = SDODataset(sdo_files, resolution=4096, patch_shape=(1024, 1024))
    sdo_dataset = StorageDataset(sdo_dataset, '%s/converted/sdo_4096' % base_path, ext_editors=[])
    sdo_dataset.convert(12)
    ###########################################################
    sdo_dataset = SDODataset(sdo_files, resolution=512)
    sdo_dataset = StorageDataset(sdo_dataset, '%s/converted/sdo_512' % base_path, ext_editors=[])
    sdo_dataset.convert(12)
    ###########################################################
    # SOHO
    ###########################################################
    soho_dataset = SOHODataset("%s/iti/soho_iti2021_prep" % base_path, months=months)
    storage_ds = StorageDataset(soho_dataset, '%s/converted/soho_1024' % base_path, ext_editors=[])
    storage_ds.convert(12)
    ###########################################################
    # STEREO
    ###########################################################
    stereo_dataset = STEREODataset("%s/iti/stereo_iti2021_prep" % base_path, months=months)
    stereo_dataset = StorageDataset(stereo_dataset, '%s/converted/stereo_1024_calibrated' % base_path,
                                    ext_editors=[])
    stereo_dataset.convert(12)



import os
import shutil

import matplotlib.pyplot as plt
import torch
from pytorch_fid.fid_score import compute_statistics_of_path, calculate_frechet_distance
from pytorch_fid.inception import InceptionV3
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm

from itipy.train.model import GeneratorAB
from itipy.trainer import skip_invalid

def calculate_fid_given_paths(paths, batch_size, device, dims):
    """Calculates the FID of two paths"""
    for p in paths:
        if not os.path.exists(p):
            raise RuntimeError('Invalid path: %s' % p)

    block_idx = InceptionV3.BLOCK_INDEX_BY_DIM[2048]

    model = InceptionV3([block_idx], normalize_input=False, resize_input=False).to(device)

    m1, s1 = compute_statistics_of_path(paths[0], model, batch_size,
                                        dims, device)
    m2, s2 = compute_statistics_of_path(paths[1], model, batch_size,
                                        dims, device)
    fid_value = calculate_frechet_distance(m1, s1, m2, s2)

    return fid_value

def computeFID(base_path, dataset_A, dataset_B, model:GeneratorAB, batch_size=4, scale_factor=1):
    path_A = os.path.join(base_path, 'A')
    path_B = os.path.join(base_path, 'B')
    path_AB = os.path.join(base_path, 'AB')

    device = torch.device('cuda' if (torch.cuda.is_available()) else 'cpu')
    model.to(device)
    model.eval()
    upsample = nn.UpsamplingBilinear2d(scale_factor=scale_factor)

    loader = DataLoader(dataset_A, batch_size=batch_size, num_workers=4)
    i, j = 0, 0
    for batch_A in tqdm(skip_invalid(loader), total=len(loader)):
        batch_A = batch_A.to(device).float()
        with torch.no_grad():
            batch_AB = model.forward(batch_A)
        for sample in batch_AB.detach().cpu().numpy():
            i += 1
            saveImage(i, path_AB, sample)
        batch_A = upsample(batch_A)
        for sample in batch_A.detach().cpu().numpy():
            j += 1
            saveImage(j, path_A, sample)

    loader = DataLoader(dataset_B, batch_size=batch_size, num_workers=4)
    i = 0
    for batch_B in tqdm(skip_invalid(loader), total=len(loader)):
        for sample in batch_B.detach().cpu().numpy():
            i += 1
            saveImage(i, path_B, sample)

    fid_AB = [calculate_fid_given_paths((os.path.join(path_B, dir), os.path.join(path_AB, dir)), batch_size, device, 2048) for dir in os.listdir(path_B)]
    fid_A = [calculate_fid_given_paths((os.path.join(path_B, dir), os.path.join(path_A, dir)), batch_size, device, 2048) for dir in os.listdir(path_B)]
    shutil.rmtree(path_A), shutil.rmtree(path_B), shutil.rmtree(path_AB) # clean up
    return fid_AB, fid_A


def saveImage(img_id, path, sample):
    for c in range(sample.shape[0]):
        os.makedirs(os.path.join(path, 'channel%d' % c), exist_ok=True)
        file_path = os.path.join(path, 'channel%d' % c, '%d.jpg' % img_id)
        plt.imsave(file_path, sample[c], vmin=-1, vmax=1, cmap='gray')





import argparse
import logging
import os

import numpy as np
from sunpy.visualization.colormaps import cm

from itipy.data.dataset import SDODataset, StorageDataset
from itipy.data.editor import SliceEditor, RandomPatchEditor, LambdaEditor
from itipy.train.model import DiscriminatorMode
from itipy.trainer import Trainer

parser = argparse.ArgumentParser(description='Train SDO EUV-To-magnetogram translations (unsigned)')
parser.add_argument('--base_dir', type=str, help='path to the results directory.')

parser.add_argument('--sdo_path', type=str, help='path to the SDO data.')
parser.add_argument('--sdo_converted_path', type=str, help='path to store the converted SDO data.')

args = parser.parse_args()
base_dir = args.base_dir

sdo_path = args.sdo_path
sdo_converted_path = args.sdo_converted_path

prediction_dir = os.path.join(base_dir, 'prediction')
os.makedirs(prediction_dir, exist_ok=True)

logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler("{0}/{1}.log".format(base_dir, "info_log")),
        logging.StreamHandler()
    ])

# Init Model
trainer = Trainer(4, 5, upsampling=0, discriminator_mode=DiscriminatorMode.CHANNELS, lambda_diversity=0,
                  norm='in_rs_aff')
trainer.cuda()


def absolute_mag(data, **kwargs):
    data[-1] = np.abs(data[-1]) * 2 - 1
    return data


abs_mag_editor = LambdaEditor(absolute_mag)

sdo_train = SDODataset(sdo_path, resolution=512, months=list(range(2, 10)), years=list(range(2011, 2021)))
sdo_euv_train = StorageDataset(sdo_train, sdo_converted_path,
                               ext_editors=[SliceEditor(0, 4), RandomPatchEditor((256, 256))])
sdo_mag_train = StorageDataset(sdo_train, sdo_converted_path,
                               ext_editors=[RandomPatchEditor((256, 256)), abs_mag_editor])

sdo_valid = SDODataset(sdo_path, resolution=512, limit=100, months=[11, 12])
sdo_euv_valid = StorageDataset(sdo_valid, sdo_converted_path,
                               ext_editors=[SliceEditor(0, 4), RandomPatchEditor((256, 256))])
sdo_mag_valid = StorageDataset(sdo_valid, sdo_converted_path,
                               ext_editors=[RandomPatchEditor((256, 256)), abs_mag_editor])

plot_settings_A = [
    {"cmap": cm.sdoaia171, "title": "SECCHI 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "SECCHI 195", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "SECCHI 284", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "SECCHI 304", 'vmin': -1, 'vmax': 1}
]
plot_settings_B = [
    {"cmap": cm.sdoaia171, "title": "AIA 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "AIA 193", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "AIA 211", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "AIA 304", 'vmin': -1, 'vmax': 1},
    {"cmap": "gray", "title": "HMI Magnetogram", 'vmin': -1, 'vmax': 1}
]

trainer.startBasicTraining(base_dir, sdo_euv_train, sdo_mag_train, sdo_euv_valid, sdo_mag_valid,
                           plot_settings_A, plot_settings_B, num_workers=4)



import argparse
import datetime
from datetime import timedelta

from itipy.download.download_hmi_continuum import HMIContinuumDownloader

parser = argparse.ArgumentParser(description='Download SDO/HMI data for video')
parser.add_argument('--download_dir', type=str, help='path to the download directory.')
parser.add_argument('--email', type=str, help='registered email address for JSOC.')
args = parser.parse_args()

hinode_date = datetime.datetime(2014, 11, 22, 1, 30)
hinode_dates = [hinode_date - timedelta(seconds=45) * i for i in range(2 * 80)]

fetcher = HMIContinuumDownloader(ds_path=args.download_dir, num_worker_threads=2,
                                 ignore_quality=True,
                                 series='hmi.Ic_45s', email=args.email)
fetcher.fetchDates(hinode_dates)



import logging
import os
from datetime import datetime
from random import randint

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
from torch.autograd import Variable
from torch.nn import InstanceNorm2d
from torch.utils.data import DataLoader

from itipy.train.model import GeneratorAB, GeneratorBA, Discriminator, NoiseEstimator, DiscriminatorMode


class Trainer(nn.Module):
    def __init__(self, input_dim_a, input_dim_b, upsampling=0, noise_dim=16, n_filters=64,
                 activation='tanh', norm='in_rs_aff', use_batch_statistic=False,
                 n_discriminators=3, discriminator_mode=DiscriminatorMode.SINGLE,
                 depth_generator=3, depth_discriminator=4, depth_noise=4, skip_connections=True,
                 lambda_discriminator=1, lambda_reconstruction=1, lambda_reconstruction_id=.1,
                 lambda_content=10, lambda_content_id=1, lambda_diversity=1, lambda_noise=1,
                 learning_rate=1e-4):
        super().__init__()

        logging.info("######################### Model Configuration ##########################")
        logging.info("Channels (A/B): %s/%s" % (str(input_dim_a), str(input_dim_b)))
        logging.info("Upsampling: %d" % upsampling)
        logging.info("Generator Depth:   %d" % depth_generator)
        logging.info("Discriminator Depth:   %d" % depth_discriminator)
        logging.info("Noise Depth:   %d" % depth_noise)
        logging.info("Number of Discriminators:   %d" % n_discriminators)
        logging.info("Discriminator Mode:   %s" % discriminator_mode)
        logging.info("Base Number of Filters:   %d" % n_filters)
        logging.info("Activation:   %s" % str(activation))
        logging.info("Normalization:   %s" % str(norm))
        logging.info("Skip Connections:   %s" % str(skip_connections))
        logging.info("Batch Statistic:   %s" % str(use_batch_statistic))
        logging.info("Learning Rate:   %f" % learning_rate)
        logging.info("Lambda Discriminator Loss:   %f" % lambda_discriminator)
        logging.info("Lambda Reconstruction Loss:   %f" % lambda_reconstruction)
        logging.info("Lambda Content Loss:   %f" % lambda_content)
        logging.info("Lambda Reconstruction ID Loss:   %f" % lambda_reconstruction_id)
        logging.info("Lambda Content ID Loss:   %f" % lambda_content_id)
        logging.info("Lambda Noise Loss:   %f" % lambda_noise)
        logging.info("Lambda Diversity Loss:   %f" % lambda_diversity)
        logging.info("START TIME:   %s" % datetime.now())
        logging.info("########################################################################")

        self.noise_dim = noise_dim
        self.input_dim_a = input_dim_a
        self.input_dim_b = input_dim_b
        ############################## MODEL CONFIG ###############################
        self.n_filters = n_filters
        self.depth_generator = depth_generator
        self.depth_discriminator = depth_discriminator
        self.depth_noise = depth_noise
        self.upsampling = upsampling

        ############################## LOSS WEIGHTS ###############################
        self.lambda_discriminator = lambda_discriminator
        self.lambda_reconstruction = lambda_reconstruction
        self.lambda_reconstruction_id = lambda_reconstruction_id
        self.lambda_content_id = lambda_content_id
        self.lambda_content = lambda_content
        self.lambda_diversity = lambda_diversity
        self.lambda_noise = lambda_noise

        ############################## INIT NETWORKS ###############################
        self.gen_ab = GeneratorAB(input_dim_a, input_dim_b, depth_generator, upsampling, n_filters,
                                  norm=norm, output_activ=activation, pad_type='reflect', skip_connections=skip_connections)  # generator for domain a-->b
        self.gen_ba = GeneratorBA(input_dim_b, input_dim_a, noise_dim, depth_generator, depth_noise, upsampling,
                                  n_filters, norm=norm, output_activ=activation,
                                  pad_type='reflect', skip_connections=skip_connections)  # generator for domain b-->a
        self.dis_a = Discriminator(input_dim_a, n_filters, n_discriminators,
                                   depth_discriminator, discriminator_mode,
                                   norm=norm, batch_statistic=use_batch_statistic)  # discriminator for domain a
        self.dis_b = Discriminator(input_dim_b, n_filters // 2 ** upsampling, n_discriminators,
                                   depth_discriminator + upsampling, discriminator_mode,
                                   norm=norm, batch_statistic=use_batch_statistic)  # discriminator for domain b
        self.estimator_noise = NoiseEstimator(input_dim_a, n_filters, noise_dim,
                                              depth_noise, norm=norm, activation='relu', pad_type='reflect')
        self.downsample = nn.AvgPool2d(2 ** upsampling)
        self.upsample = nn.UpsamplingBilinear2d(scale_factor=2 ** upsampling)
        self.module_list = nn.ModuleList([self.gen_ab, self.gen_ba, self.dis_a, self.dis_b, self.estimator_noise])

        # Setup the optimizers
        dis_params = list(self.dis_a.parameters()) + list(self.dis_b.parameters())
        gen_params = list(self.gen_ab.parameters()) + list(self.gen_ba.parameters()) + list(
            self.estimator_noise.parameters())
        self.dis_opt = torch.optim.Adam(dis_params, lr=learning_rate, betas=(0.5, 0.9))
        self.gen_opt = torch.optim.Adam(gen_params, lr=learning_rate, betas=(0.5, 0.9))

        # Training utils
        self.gen_stack = []
        loss_keys = [
            'iteration',
            'loss_gen_a_identity',
            'loss_gen_b_identity',
            'loss_gen_a_translate',
            'loss_gen_b_translate',
            'loss_gen_adv_a',
            'loss_gen_adv_b',
            'loss_gen_a_content',
            'loss_gen_b_content',
            'loss_gen_a_identity_content',
            'loss_gen_b_identity_content',
            'loss_gen_aba_noise',
            'loss_gen_a_identity_noise',
            'loss_dis_a',
            'loss_dis_b',
            'loss_gen_diversity']
        self.train_loss = {key: [] for key in loss_keys}
        self.valid_loss = {key: [] for key in loss_keys}

        logging.info("Total Parameters Generator (BA/AB): %.02f/%.02f M" %
                     (sum(p.numel() for p in self.gen_ba.parameters()) / 1e6,
                      sum(p.numel() for p in self.gen_ab.parameters()) / 1e6))
        logging.info("Total Parameters Discriminator (A/B): %.02f/%.02f M" %
                     (sum(p.numel() for p in self.dis_a.parameters()) / 1e6,
                      sum(p.numel() for p in self.dis_b.parameters()) / 1e6))
        logging.info("Total Parameters Noise Estimator: %.02f M" %
                     (sum(p.numel() for p in self.estimator_noise.parameters()) / 1e6))

    def recon_criterion(self, input, target):
        return torch.mean(torch.abs(input - target))

    def forward(self, x_a, x_b):
        n_gen = self.generateNoise(x_b)
        x_ab = self.gen_ab(x_a)
        x_ba = self.gen_ba(x_b, n_gen)
        return x_ab, x_ba

    def forwardAB(self, x_a):
        x_ab = self.gen_ab(x_a)
        return x_ab

    def forwardBA(self, x_b):
        n_gen = self.generateNoise(x_b)
        x_ba = self.gen_ba(x_b, n_gen)
        return x_ba

    def forwardABA(self, x_a):
        n_a = self.estimator_noise(x_a)
        x_ab = self.gen_ab(x_a)
        x_aba = self.gen_ba(x_ab, n_a)
        return x_ab, x_aba

    def forwardBAB(self, x_b):
        n_gen = self.generateNoise(x_b)
        x_ba = self.gen_ba(x_b, n_gen)
        x_bab = self.gen_ab(x_ba)
        return x_ba, x_bab

    def generator_update(self, x_a, x_b):
        self.gen_opt.zero_grad()

        # noise init
        n_a = self.estimator_noise(x_a)
        n_gen = self.generateNoise(x_b)

        # identity
        if self.input_dim_a > self.input_dim_b:
            x_b_downsample = self.downsample(x_b)
            x_b_downsample = torch.repeat_interleave(x_b_downsample, repeats=self.input_dim_a, dim=1)
            x_b_identity = self.gen_ab(x_b_downsample)

            x_a_upsample = self.upsample(x_a)
            idx = randint(0, self.input_dim_a - 1)
            x_a_upsample = x_a_upsample[:, idx:idx + 1]
            x_a_identity = self.gen_ba(x_a_upsample, n_a)
        else:  # channel difference 0 == no change of dims
            c_diff = self.input_dim_b - self.input_dim_a

            x_b_downsample = self.downsample(x_b)
            x_b_downsample = x_b_downsample[:, :-c_diff] if c_diff > 0 else x_b_downsample
            x_b_identity = self.gen_ab(x_b_downsample)

            x_a_upsample = self.upsample(x_a)
            x_a_upsample = F.pad(x_a_upsample, [0, 0, 0, 0, 0, c_diff], mode='constant',
                                 value=0) if c_diff > 0 else x_a_upsample
            x_a_identity = self.gen_ba(x_a_upsample, n_a)

        n_a_identity = self.estimator_noise(x_a_identity)

        # translate 1
        x_ab = self.gen_ab(x_a)
        x_ba = self.gen_ba(x_b, n_gen)
        # noise 1
        n_ba = self.estimator_noise(x_ba)

        # translate 2
        x_aba = self.gen_ba(x_ab, n_a)
        x_bab = self.gen_ab(x_ba)
        # noise 2
        n_aba = self.estimator_noise(x_aba)

        # reconstruction loss
        self.loss_gen_a_identity = self.recon_criterion(x_a_identity, x_a)
        self.loss_gen_b_identity = self.recon_criterion(x_b_identity, x_b)
        self.loss_gen_a_translate = self.recon_criterion(x_aba, x_a)
        self.loss_gen_b_translate = self.recon_criterion(x_bab, x_b)
        # GAN loss
        self.loss_gen_adv_a = self.dis_a.calc_gen_loss(x_ba)
        self.loss_gen_adv_b = self.dis_b.calc_gen_loss(x_ab)
        # Content loss
        self.loss_gen_a_content = torch.mean(self.dis_a.calc_content_loss(x_a, x_aba))
        self.loss_gen_b_content = torch.mean(self.dis_b.calc_content_loss(x_b, x_bab))
        self.loss_gen_a_identity_content = torch.mean(self.dis_a.calc_content_loss(x_a, x_a_identity))
        self.loss_gen_b_identity_content = torch.mean(self.dis_b.calc_content_loss(x_b, x_b_identity))
        # Noise loss
        self.loss_gen_ba_noise = self.recon_criterion(n_ba, n_gen)
        self.loss_gen_aba_noise = self.recon_criterion(n_aba, n_a)
        self.loss_gen_a_identity_noise = self.recon_criterion(n_a_identity, n_a)
        # Diversity loss
        if self.lambda_diversity > 0:
            with torch.no_grad():  # no double back-prop
                n_gen_2 = self.generateNoise(x_b)
                x_ba_div = self.gen_ba(x_b, n_gen_2).detach()
            diversity_diff = self.dis_a.calc_content_loss(x_ba, x_ba_div)
            self.loss_gen_diversity = torch.mean(
                - torch.log((diversity_diff + 1e-6) / (torch.mean(torch.abs(n_gen - n_gen_2), [1, 2, 3]) + 1e-6)))
        else:
            self.loss_gen_diversity = torch.tensor(0.0)
        # total loss
        self.loss_gen_total = self.lambda_reconstruction_id * (self.loss_gen_a_identity + self.loss_gen_b_identity) + \
                              self.lambda_reconstruction * (self.loss_gen_a_translate + self.loss_gen_b_translate) + \
                              self.lambda_discriminator * (self.loss_gen_adv_a + self.loss_gen_adv_b) + \
                              self.lambda_content * (self.loss_gen_a_content + self.loss_gen_b_content) + \
                              self.lambda_content_id * (
                                      self.loss_gen_a_identity_content + self.loss_gen_b_identity_content) + \
                              self.lambda_noise * (
                                      self.loss_gen_ba_noise + self.loss_gen_aba_noise + self.loss_gen_a_identity_noise) + \
                              self.lambda_diversity * self.loss_gen_diversity
        self.loss_gen_total.backward()
        self.gen_opt.step()

    def validate(self, x_a, x_b):
        self.eval()
        with torch.no_grad():
            # noise init
            n_a = self.estimator_noise(x_a)
            n_gen = self.generateNoise(x_b)

            # identity
            x_b_identity = self.gen_ab(self.downsample(x_b))
            x_a_identity = self.gen_ba(self.upsample(x_a), n_a)
            n_a_identity = self.estimator_noise(x_a_identity)

            # translate 1
            x_ab = self.gen_ab(x_a)
            x_ba = self.gen_ba(x_b, n_gen)

            # translate 2
            x_aba = self.gen_ba(x_ab, n_a)
            x_bab = self.gen_ab(x_ba)
            # noise 2
            n_aba = self.estimator_noise(x_aba)

            # reconstruction loss
            self.valid_loss_gen_a_identity = self.recon_criterion(x_a_identity, x_a)
            self.valid_loss_gen_b_identity = self.recon_criterion(x_b_identity, x_b)
            self.valid_loss_gen_a_translate = self.recon_criterion(x_aba, x_a)
            self.valid_loss_gen_b_translate = self.recon_criterion(x_bab, x_b)
            # GAN loss
            self.valid_loss_gen_adv_a = self.dis_a.calc_gen_loss(x_ba)
            self.valid_loss_gen_adv_b = self.dis_b.calc_gen_loss(x_ab)
            self.valid_loss_dis_a = self.dis_a.calc_gen_loss(x_a)  # use only real images fro validation
            self.valid_loss_dis_b = self.dis_b.calc_gen_loss(x_b)
            # Content loss
            self.valid_loss_gen_a_content = torch.mean(self.dis_a.calc_content_loss(x_a, x_aba))
            self.valid_loss_gen_b_content = torch.mean(self.dis_b.calc_content_loss(x_b, x_bab))
            self.valid_loss_gen_a_identity_content = torch.mean(self.dis_a.calc_content_loss(x_a, x_a_identity))
            self.valid_loss_gen_b_identity_content = torch.mean(self.dis_b.calc_content_loss(x_b, x_b_identity))
            # Noise loss
            self.valid_loss_gen_aba_noise = self.recon_criterion(n_aba, n_a)
            self.valid_loss_gen_a_identity_noise = self.recon_criterion(n_a_identity, n_a)
        self.train()

    def fill_stack(self, x_a, x_b):
        self.eval()
        with torch.no_grad():
            n_gen = self.generateNoise(x_b)
            x_ab = self.gen_ab(x_a).cpu().detach()
            x_ba = self.gen_ba(x_b, n_gen).cpu().detach()
            self.gen_stack.append((x_ab, x_ba))
        self.train()

    def discriminator_update(self, x_a, x_b):
        self.dis_opt.zero_grad()
        # noise
        n_gen = self.generateNoise(x_b)
        # translate
        with torch.no_grad():
            x_ab = self.gen_ab(x_a).cpu().detach()
            x_ba = self.gen_ba(x_b, n_gen).cpu().detach()
            self.gen_stack.append((x_ab, x_ba))
            x_ab, x_ba = self.gen_stack.pop(0)
            x_ab, x_ba = x_ab.cuda(), x_ba.cuda()

        # D loss
        self.loss_dis_a = self.dis_a.calc_dis_loss(x_ba, x_a)
        self.loss_dis_b = self.dis_b.calc_dis_loss(x_ab, x_b)
        self.loss_dis_total = self.loss_dis_a + self.loss_dis_b
        self.loss_dis_total.backward()
        self.dis_opt.step()

    def generateNoise(self, x_b):
        n_gen = Variable(torch.rand(x_b.size(0), self.noise_dim,
                                    x_b.size(2) // 2 ** (self.depth_noise + self.upsampling),
                                    x_b.size(3) // 2 ** (self.depth_noise + self.upsampling)).cuda())
        return n_gen

    def resume(self, checkpoint_dir, epoch=None):
        path = os.path.join(checkpoint_dir, 'checkpoint.pt') if epoch is None else os.path.join(checkpoint_dir,
                                                                                                'checkpoint_%d.pt' % epoch)
        if not os.path.exists(path):
            return 0
        state_dict = torch.load(path)
        # Load generators
        self.gen_ab.load_state_dict(state_dict['gen_ab'])
        self.gen_ba.load_state_dict(state_dict['gen_ba'])
        self.estimator_noise.load_state_dict(state_dict['noise_est'])
        # Load discriminators
        self.dis_a.load_state_dict(state_dict['disc_a'])
        self.dis_b.load_state_dict(state_dict['disc_b'])
        # Load optimizers
        self.gen_opt.load_state_dict(state_dict['opt_gen'])
        self.dis_opt.load_state_dict(state_dict['opt_dis'])
        # Load history
        last_iteration = state_dict['iteration']
        self.train_loss = state_dict['train_loss'] if 'train_loss' in state_dict else self.train_loss
        self.valid_loss = state_dict['valid_loss'] if 'valid_loss' in state_dict else self.valid_loss
        print('Resume from iteration %d' % last_iteration)
        return last_iteration

    def save(self, checkpoint_dir, iterations):
        # Save generators, discriminators, and optimizers
        state_path = os.path.join(checkpoint_dir, 'checkpoint.pt')
        checkpoint_path = os.path.join(checkpoint_dir, 'checkpoint_%05d.pt' % (iterations + 1))
        state = {'iteration': iterations + 1,
                 'gen_ab': self.gen_ab.state_dict(),
                 'gen_ba': self.gen_ba.state_dict(),
                 'noise_est': self.estimator_noise.state_dict(),
                 'disc_a': self.dis_a.state_dict(),
                 'disc_b': self.dis_b.state_dict(),
                 'opt_gen': self.gen_opt.state_dict(),
                 'opt_dis': self.dis_opt.state_dict(),
                 'train_loss': self.train_loss,
                 'valid_loss': self.valid_loss}
        torch.save(state, state_path)
        if (iterations + 1) % 20000 == 0:
            torch.save(state, checkpoint_path)
            torch.save(self.gen_ab, os.path.join(checkpoint_dir, 'generator_AB.pt'))
            torch.save(self.gen_ba, os.path.join(checkpoint_dir, 'generator_BA.pt'))

    def updateMomentum(self, momentum):
        for module in self.modules():
            if isinstance(module, InstanceNorm2d):
                module.momentum = momentum

    def startBasicTraining(self, base_dir, ds_A, ds_B, ds_valid_A=None, ds_valid_B=None,
                           plot_settings_A=None, plot_settings_B=None, additional_callbacks=[],
                           iterations=int(1e8), num_workers=8, validation_history=False, batch_size=1):
        self.cuda()
        start_it = self.resume(base_dir)
        # Init Callbacks
        from itipy.callback import HistoryCallback, ProgressCallback, SaveCallback, PlotBAB, PlotABA, ValidationHistoryCallback
        history_callback = HistoryCallback(self, base_dir)
        progress_callback = ProgressCallback(self)
        save_callback = SaveCallback(self, base_dir)
        callbacks = [history_callback, progress_callback, save_callback]
        if ds_valid_B is not None or ds_valid_A is not None:
            prediction_dir = os.path.join(base_dir, 'prediction')
            os.makedirs(prediction_dir, exist_ok=True)
        if ds_valid_B is not None:
            callbacks += [PlotBAB(ds_valid_B.sample(4), self, prediction_dir, log_iteration=1000,
                                  plot_settings_A=plot_settings_A, plot_settings_B=plot_settings_B)]
        if ds_valid_A is not None:
            callbacks += [PlotABA(ds_valid_A.sample(4), self, prediction_dir, log_iteration=1000,
                                  plot_settings_A=plot_settings_A, plot_settings_B=plot_settings_B)]
        if validation_history:
            assert ds_valid_B is not None and ds_valid_A is not None, 'Validation history requires validation data sets!'
            callbacks += [ValidationHistoryCallback(self, ds_valid_A, ds_valid_B, base_dir, 1000)]
        callbacks += additional_callbacks
        # init data loaders
        B_iterator = loop(DataLoader(ds_B, batch_size=batch_size, shuffle=True, num_workers=num_workers))
        A_iterator = loop(DataLoader(ds_A, batch_size=batch_size, shuffle=True, num_workers=num_workers))
        # start update cycle
        for it in range(start_it, iterations):
            self.train()
            if it > 100000: # fix running stats
                self.gen_ab.eval()
                self.gen_ba.eval()
            x_a, x_b = next(A_iterator), next(B_iterator)
            x_a, x_b = x_a.float().cuda().detach(), x_b.float().cuda().detach()
            self.discriminator_update(x_a, x_b)
            #
            x_a, x_b = next(A_iterator), next(B_iterator)
            x_a, x_b = x_a.float().cuda().detach(), x_b.float().cuda().detach()
            self.generator_update(x_a, x_b)
            torch.cuda.synchronize()
            #
            self.eval()
            with torch.no_grad():
                for callback in callbacks:
                    callback(it)


def convertSet(data_set, store_path):
    loader = DataLoader(data_set, batch_size=1, shuffle=False, num_workers=16)
    for i, sample in enumerate(loader):
        np.save(os.path.join(store_path, '%d.npy' % i), sample[0])


def loop(iterable):
    it = iterable.__iter__()
    #
    while True:
        try:
            yield it.next()
        except StopIteration:
            it = iterable.__iter__()
            yield it.next()
        except Exception as ex:
            logging.error(str(ex))
            continue


def skip_invalid(iterable):
    it = iterable.__iter__()
    #
    while True:
        try:
            yield next(it)
        except StopIteration as ex:
            return
        except (AssertionError, ValueError, Exception) as ex:
            logging.error(str(ex))
            continue



import gc
import glob
import logging
import os
import random
import warnings
from collections import Iterable
from enum import Enum
from typing import List, Union

import numpy as np
from astropy.visualization import AsinhStretch
from dateutil.parser import parse
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm

from itipy.data.editor import Editor, LoadMapEditor, KSOPrepEditor, NormalizeRadiusEditor, \
    MapToDataEditor, ImageNormalizeEditor, ReshapeEditor, sdo_norms, NormalizeEditor, \
    AIAPrepEditor, RemoveOffLimbEditor, StackEditor, soho_norms, NanEditor, LoadFITSEditor, \
    KSOFilmPrepEditor, ScaleEditor, ExpandDimsEditor, FeaturePatchEditor, EITCheckEditor, NormalizeExposureEditor, \
    PassEditor, BrightestPixelPatchEditor, stereo_norms, LimbDarkeningCorrectionEditor, hinode_norms, gregor_norms, \
    LoadGregorGBandEditor, DistributeEditor, RecenterEditor, AddRadialDistanceEditor, SECCHIPrepEditor, \
    SOHOFixHeaderEditor, PaddingEditor


class Norm(Enum):
    CONTRAST = 'contrast'
    IMAGE = 'image'
    PEAK = 'adjusted'
    NONE = 'none'

class ArrayDataset(Dataset):

    def __init__(self, data, editors: List[Editor], **kwargs):
        self.data = data
        self.editors = editors

        super().__init__()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        data, _ = self.getIndex(idx)
        return data

    def sample(self, n_samples):
        it = DataLoader(self, batch_size=1, shuffle=True, num_workers=4).__iter__()
        samples = []
        while len(samples) < n_samples:
            try:
                samples.append(next(it).detach().numpy()[0])
            except Exception as ex:
                logging.error(str(ex))
                continue
        del it
        return np.array(samples)

    def getIndex(self, idx):
        try:
            return self.convertData(self.data[idx])
        except Exception as ex:
            logging.error('Unable to convert %s: %s' % (self.data[idx], ex))
            raise ex

    def getId(self, idx):
        return str(idx)

    def convertData(self, data):
        kwargs = {}
        for editor in self.editors:
            data, kwargs = editor.convert(data, **kwargs)
        return data, kwargs

    def addEditor(self, editor):
        self.editors.append(editor)


class BaseDataset(Dataset):

    def __init__(self, data: Union[str, list], editors: List[Editor], ext: str = None, limit: int = None,
                 months: list = None, date_parser=None, **kwargs):
        if isinstance(data, str):
            pattern = '*' if ext is None else '*' + ext
            data = sorted(glob.glob(os.path.join(data, "**", pattern), recursive=True))
        assert isinstance(data, Iterable), 'Dataset requires list of samples or path to files!'
        if months:  # assuming filename is parsable datetime
            if date_parser is None:
                date_parser = lambda f: parse(os.path.basename(f).split('.')[0])
            data = [d for d in data if date_parser(d).month in months]
        if limit is not None:
            data = random.sample(list(data), limit)
        self.data = data
        self.editors = editors

        super().__init__()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        data, _ = self.getIndex(idx)
        return data

    def sample(self, n_samples):
        it = DataLoader(self, batch_size=1, shuffle=True, num_workers=4).__iter__()
        samples = []
        while len(samples) < n_samples:
            try:
                samples.append(next(it).detach().numpy()[0])
            except Exception as ex:
                logging.error(str(ex))
                continue
        del it
        return np.array(samples)

    def getIndex(self, idx):
        try:
            return self.convertData(self.data[idx])
        except Exception as ex:
            logging.error('Unable to convert %s: %s' % (self.data[idx], ex))
            raise ex

    def getId(self, idx):
        return os.path.basename(self.data[idx]).split('.')[0]

    def convertData(self, data):
        kwargs = {}
        for editor in self.editors:
            data, kwargs = editor.convert(data, **kwargs)
        return data, kwargs

    def addEditor(self, editor):
        self.editors.append(editor)


class StackDataset(BaseDataset):

    def __init__(self, data_sets, limit=None, **kwargs):
        self.data_sets = data_sets

        editors = [StackEditor(data_sets)]
        super().__init__(list(range(len(data_sets[0]))), editors, limit=limit)

    def getId(self, idx):
        return os.path.basename(self.data_sets[0].data[idx]).split('.')[0]


class StorageDataset(Dataset):
    def __init__(self, dataset: BaseDataset, store_dir, ext_editors=[]):
        self.dataset = dataset
        self.store_dir = store_dir
        self.ext_editors = ext_editors
        os.makedirs(store_dir, exist_ok=True)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        id = self.dataset.getId(idx)
        store_path = os.path.join(self.store_dir, '%s.npy' % id)
        if os.path.exists(store_path):
            data = np.load(store_path, mmap_mode='r')
            data = self.convertData(data)
            return data
        data = self.dataset[idx]
        np.save(store_path, data)
        data = self.convertData(data)
        return data

    def convertData(self, data):
        kwargs = {}
        for editor in self.ext_editors:
            data, kwargs = editor.convert(data, **kwargs)
        return data

    def sample(self, n_samples):
        it = DataLoader(self, batch_size=1, shuffle=True, num_workers=4).__iter__()
        samples = []
        while len(samples) < n_samples:
            try:
                samples.append(next(it).detach().numpy())
            except Exception as ex:
                logging.error(str(ex))
                continue
        del it
        return np.concatenate(samples)

    def convert(self, n_worker):
        it = DataLoader(self, batch_size=1, shuffle=False, num_workers=n_worker).__iter__()
        for i in tqdm(range(len(self.dataset))):
            try:
                next(it)
                gc.collect()
            except StopIteration:
                return
            except Exception as ex:
                logging.error('Invalid data: %s' % self.dataset.data[i])
                logging.error(str(ex))
                continue


def get_intersecting_files(path, dirs, months=None, years=None, n_samples=None, ext=None, basenames=None, **kwargs):
    pattern = '*' if ext is None else '*' + ext
    if basenames is None:
        basenames = [[os.path.basename(path) for path in glob.glob(os.path.join(path, str(d), '**', pattern), recursive=True)] for d in dirs]
        basenames = list(set(basenames[0]).intersection(*basenames))
    if months:  # assuming filename is parsable datetime
        basenames = [bn for bn in basenames if parse(bn.split('.')[0]).month in months]
    if years:  # assuming filename is parsable datetime
        basenames = [bn for bn in basenames if parse(bn.split('.')[0]).year in years]
    basenames = sorted(list(basenames))
    if n_samples:
        basenames = basenames[::len(basenames) // n_samples]
    return [[os.path.join(path, str(dir), b) for b in basenames] for dir in dirs]


class SDODataset(StackDataset):

    def __init__(self, data, patch_shape=None, resolution=2048, ext='.fits', **kwargs):
        if isinstance(data, list):
            paths = data
        else:
            paths = get_intersecting_files(data, ['171', '193', '211', '304', '6173'], ext=ext, **kwargs)
        data_sets = [AIADataset(paths[0], 171, resolution=resolution, **kwargs),
                     AIADataset(paths[1], 193, resolution=resolution, **kwargs),
                     AIADataset(paths[2], 211, resolution=resolution, **kwargs),
                     AIADataset(paths[3], 304, resolution=resolution, **kwargs),
                     HMIDataset(paths[4], 'mag', resolution=resolution)
                     ]
        super().__init__(data_sets, **kwargs)
        if patch_shape is not None:
            self.addEditor(BrightestPixelPatchEditor(patch_shape))


class SOHODataset(StackDataset):

    def __init__(self, data, patch_shape=None, resolution=1024, ext='.fits', wavelengths=None, **kwargs):
        wavelengths = [171, 195, 284, 304, 'mag', ] if wavelengths is None else wavelengths
        if isinstance(data, list):
            paths = data
        else:
            paths = get_intersecting_files(data, wavelengths, ext=ext, **kwargs)

        ds = {171: EITDataset, 195: EITDataset, 284: EITDataset, 304: EITDataset, 'mag': MDIDataset}
        data_sets = [ds[wl_id](files, wavelength=wl_id, resolution=resolution, ext=ext)
                     for wl_id, files in zip(wavelengths, paths)]

        super().__init__(data_sets, **kwargs)
        if patch_shape is not None:
            self.addEditor(BrightestPixelPatchEditor(patch_shape))


class STEREODataset(StackDataset):

    def __init__(self, data, patch_shape=None, resolution=1024, **kwargs):
        if isinstance(data, list):
            paths = data
        else:
            paths = get_intersecting_files(data, ['171', '195', '284', '304'], **kwargs)
        data_sets = [SECCHIDataset(paths[0], 171, resolution=resolution),
                     SECCHIDataset(paths[1], 195, resolution=resolution),
                     SECCHIDataset(paths[2], 284, resolution=resolution),
                     SECCHIDataset(paths[3], 304, resolution=resolution, degradation=[-9.42497209e-05, 2.27153104e+00]),
                     ]
        super().__init__(data_sets, **kwargs)
        if patch_shape is not None:
            self.addEditor(BrightestPixelPatchEditor(patch_shape))


class EITDataset(BaseDataset):

    def __init__(self, data, wavelength, resolution=1024, ext='.fits', **kwargs):
        norm = soho_norms[wavelength]

        editors = [LoadMapEditor(),
                   EITCheckEditor(),
                   SOHOFixHeaderEditor(),
                   NormalizeRadiusEditor(resolution),
                   MapToDataEditor(),
                   NormalizeEditor(norm),
                   ReshapeEditor((1, resolution, resolution))]
        super().__init__(data, editors=editors, ext=ext, **kwargs)


class MDIDataset(BaseDataset):

    def __init__(self, data, resolution=1024, ext='.fits', **kwargs):
        norm = soho_norms[6173]
        editors = [LoadMapEditor(),
                   SOHOFixHeaderEditor(),
                   NormalizeRadiusEditor(resolution),
                   RemoveOffLimbEditor(),
                   MapToDataEditor(),
                   NanEditor(),
                   NormalizeEditor(norm),
                   ReshapeEditor((1, resolution, resolution))]
        super().__init__(data, editors=editors, ext=ext, **kwargs)


class AIADataset(BaseDataset):

    def __init__(self, data, wavelength, resolution=2048, ext='.fits', calibration='auto', **kwargs):
        norm = sdo_norms[wavelength]

        editors = [LoadMapEditor(),
                   NormalizeRadiusEditor(resolution),
                   AIAPrepEditor(calibration=calibration),
                   MapToDataEditor(),
                   NormalizeEditor(norm),
                   ReshapeEditor((1, resolution, resolution))]
        super().__init__(data, editors=editors, ext=ext, **kwargs)


class HMIDataset(BaseDataset):

    def __init__(self, data, id, resolution=2048, ext='.fits', **kwargs):
        norm = sdo_norms[id]

        editors = [LoadMapEditor(),
                   NormalizeRadiusEditor(resolution),
                   RemoveOffLimbEditor(),
                   MapToDataEditor(),
                   PaddingEditor((resolution, resolution)),  # fix field-of-view of subframe
                   NanEditor(),
                   NormalizeEditor(norm),
                   ReshapeEditor((1, resolution, resolution))]
        super().__init__(data, editors=editors, ext=ext, **kwargs)


class HMIContinuumDataset(BaseDataset):

    def __init__(self, data, patch_shape=None, **kwargs):
        norm = sdo_norms['continuum']

        editors = [LoadMapEditor(),
                   RecenterEditor(),
                   ScaleEditor(0.6),
                   FeaturePatchEditor(patch_shape) if patch_shape is not None else PassEditor(),
                   MapToDataEditor(),
                   NanEditor(),
                   NormalizeEditor(norm),
                   ExpandDimsEditor()]
        super().__init__(data, editors=editors, **kwargs)


class SDOMLDataset(BaseDataset):

    def __init__(self, data, wavelength, resolution=2048, ext='.fits', **kwargs):
        norm = sdo_norms[wavelength]

        editors = [LoadMapEditor(),
                   NormalizeRadiusEditor(resolution),
                   AIAPrepEditor(),
                   MapToDataEditor(),
                   NormalizeEditor(norm),
                   ReshapeEditor((1, resolution, resolution))]
        super().__init__(data, editors=editors, ext=ext, **kwargs)

class HinodeDataset(BaseDataset):

    def __init__(self, data, scale=0.15, wavelength='continuum', **kwargs):
        norm = hinode_norms[wavelength]

        editors = [LoadMapEditor(),
                   ScaleEditor(scale),
                   NormalizeExposureEditor(),
                   MapToDataEditor(),
                   NanEditor(),
                   NormalizeEditor(norm),
                   ExpandDimsEditor()]
        super().__init__(data, editors=editors, **kwargs)


class SECCHIDataset(BaseDataset):

    def __init__(self, data, wavelength, resolution=1024, degradation=None,**kwargs):
        norm = stereo_norms[wavelength]

        editors = [LoadMapEditor(),
                   SECCHIPrepEditor(degradation),
                   NormalizeRadiusEditor(resolution),
                   MapToDataEditor(),
                   NormalizeEditor(norm),
                   ReshapeEditor((1, resolution, resolution))]
        super().__init__(data, editors=editors, **kwargs)


class KSODataset(BaseDataset):

    def __init__(self, data: Union[str, list], resolution=256, ext=".fts.gz", **kwargs):
        editors = [LoadMapEditor(),
                   KSOPrepEditor(),
                   NormalizeRadiusEditor(resolution),
                   MapToDataEditor(),
                   ImageNormalizeEditor(0, 1000),
                   ReshapeEditor((1, resolution, resolution))]
        super().__init__(data, editors=editors, ext=ext, **kwargs)


class KSOFlatDataset(BaseDataset):

    def __init__(self, data, resolution=256, ext=".fts.gz", date_parser=None, **kwargs):
        editors = [LoadMapEditor(),
                   KSOPrepEditor(),
                   NormalizeRadiusEditor(resolution, 0),
                   LimbDarkeningCorrectionEditor(),
                   MapToDataEditor(),
                   ImageNormalizeEditor(0.65, 1.5, stretch=AsinhStretch(0.5)),
                   NanEditor(-1),
                   ReshapeEditor((1, resolution, resolution))]
        if date_parser is None:
            date_parser = lambda f: parse(os.path.basename(f)[14:-7].replace('_', 'T'))
        super().__init__(data, editors=editors, ext=ext, date_parser=date_parser, **kwargs)


class KSOFilmDataset(BaseDataset):

    def __init__(self, data, resolution=256, ext=".fts.gz", date_parser=None, **kwargs):
        editors = [LoadFITSEditor(),
                   KSOFilmPrepEditor(),
                   NormalizeRadiusEditor(resolution, 0),
                   LimbDarkeningCorrectionEditor(),
                   MapToDataEditor(),
                   ImageNormalizeEditor(0.39, 1.94, stretch=AsinhStretch(0.5)),
                   NanEditor(-1),
                   ReshapeEditor((1, resolution, resolution))]
        if date_parser is None:
            date_parser = lambda f: parse(os.path.basename(f)[-22:-7].replace('_', 'T'))
        super().__init__(data, editors=editors, ext=ext, date_parser=date_parser, **kwargs)


class GregorDataset(BaseDataset):

    def __init__(self, data, ext='.fts', **kwargs):
        norm = gregor_norms['gband']
        sub_editors = [MapToDataEditor(),
                       NanEditor(),
                       NormalizeEditor(norm),
                       ExpandDimsEditor()]
        editors = [LoadGregorGBandEditor(), DistributeEditor(sub_editors)]

        super().__init__(data, editors, ext=ext, **kwargs)


class ZerosDataset(Dataset):

    def __init__(self, length, resolution=1024, **kwargs):
        self.shape = (1, resolution, resolution)
        self.length = length

        super().__init__()

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        return np.zeros(self.shape)



import argparse
import logging
import os
import shutil
from datetime import timedelta, datetime
from glob import glob
from multiprocessing import Pool
from random import sample

import drms
from astropy import units as u
from dateutil.relativedelta import relativedelta
from sunpy.map import Map
from sunpy.net import Fido, attrs as a
from tqdm import tqdm


class STEREODownloader:

    def __init__(self, base_path):
        self.base_path = base_path
        self.wavelengths = [304, 284, 195, 171]
        self.dirs = ['304', '284', '195', '171']
        [os.makedirs(os.path.join(base_path, dir), exist_ok=True) for dir in self.dirs]

    def downloadDate(self, sample):
        date, source = sample
        files = []
        try:
            # Download SECCHI
            for wl in self.wavelengths:
                files += [self.downloadSECCHI(date, wl, source)]
            logging.info('Download complete %s' % date.isoformat())
        except Exception as ex:
            logging.error('Unable to download %s: %s' % (date.isoformat(), str(ex)))
            [os.remove(f) for f in files if os.path.exists(f)]

    def downloadSECCHI(self, query_date, wl, source):
        file_path = os.path.join(self.base_path, str(wl), "%s.fits" % query_date.isoformat("T", timespec='seconds'))
        if os.path.exists(file_path):
            return file_path  # skip existing downloads (e.g. retry)
        #
        search = Fido.search(a.Time(query_date - timedelta(minutes=15), query_date + timedelta(minutes=15)),
                             a.Provider('SSC'), a.Source(source), a.Instrument('SECCHI'), a.Wavelength(wl * u.AA))
        assert search.file_num > 0, "No data found for %s (%s)" % (query_date.isoformat(), wl)
        search = sorted(search['vso'], key=lambda x: abs(x['Start Time'].datetime - query_date).total_seconds())
        for entry in search:
            files = Fido.fetch(entry, path=self.base_path, progress=False)
            if len(files) != 1:
                continue
            file = files[0]
            try:
                header = Map(file).meta
            except:
                os.remove(file)
                continue
            if header['NAXIS1'] < 2048 or header['NAXIS2'] < 2048 or header['NMISSING'] != 0:
                os.remove(file)
                continue
            shutil.move(file, file_path)
            return file_path

        raise Exception("No valid file found for %s (%s)!" % (query_date.isoformat(), wl))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Download STEREO data')
    parser.add_argument('--download_dir', type=str, help='path to the download directory.')
    parser.add_argument('--n_workers', type=str, help='number of parallel threads.', required=False, default=4)

    args = parser.parse_args()
    base_path = args.download_dir
    n_workers = args.n_workers

    drms_client = drms.Client(email='robert.jarolim@uni-graz.at', verbose=False)
    download_util = STEREODownloader(base_path)
    logging.basicConfig(
        level=logging.INFO,
        handlers=[
            logging.StreamHandler()
        ])

    existing_dates = set(
        [os.path.basename(f).replace('.fits', '')[:10] for f in glob('/localdata/USER/rja/stereo_iti2021/**/*.fits')])

    start_date = datetime(2008, 5, 1, 0, 0)
    end_date = datetime.now()
    num_months = (end_date.year - start_date.year) * 12 + (end_date.month - start_date.month)
    month_dates = [start_date + relativedelta(months=i) for i in range(num_months)]
    for date in month_dates:
        samples = []
        for i in range(((date + relativedelta(months=1)) - date).days):
            search_date = date + timedelta(days=i)
            if '%04d-%02d-%02d' % (search_date.year, search_date.month, search_date.day) in existing_dates:
                continue
            search = Fido.search(a.Time(search_date, search_date + timedelta(minutes=60)),
                                 a.Provider('SSC'), a.Instrument('SECCHI'), a.Wavelength(171 * u.AA))
            if search.file_num == 0:
                continue
            dates = search['vso']['Start Time']
            sources = search['vso']['Source']
            samples += [sample([(d.datetime, s) for d, s in zip(dates, sources)], 1)[0]]

        logging.info("TOTAL DAYS (%s): %d" % (date.isoformat(), len(samples)))
        with Pool(n_worker) as p:
            [None for _ in tqdm(p.imap_unordered(download_util.downloadDate, samples), total=len(samples))]



import argparse
import glob
import os
import shutil
from datetime import timedelta, datetime

import drms
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

parser = argparse.ArgumentParser(description='Download SDO data from JSOC')
parser.add_argument('--download_dir', type=str, help='path to the download directory.')
parser.add_argument('--email', type=str, help='registered email address for JSOC.')
parser.add_argument('--channels', '-channels', nargs='+', required=False,
                    default=['171', '193', '211', '304', '6173'],
                    help='subset of channels to load. The order must match the input channels of the model.')


args = parser.parse_args()
download_dir = args.download_dir
channels = args.channels
euv_channels = [c for c in channels if c != '6173']

[os.makedirs(os.path.join(download_dir, str(c)), exist_ok=True) for c in channels]

client = drms.Client(verbose=True, email=args.email)


def round_date(t):
    if t.minute >= 30:
        return t.replace(second=0, microsecond=0, minute=0) + timedelta(hours=1)
    else:
        return t.replace(second=0, microsecond=0, minute=0)


def download(ds):
    r = client.export(ds, method='url-tar', protocol='fits')
    r.wait()
    download_result = r.download(download_dir)
    for f in download_result.download:
        shutil.unpack_archive(f, os.path.join(download_dir))
        os.remove(f)
    for f in glob.glob(os.path.join(download_dir, '*.fits')):
        f_info = os.path.basename(f).split('.')
        channel = f_info[3]
        if f_info[0] == 'hmi':
            channel = '6173'
            date = round_date(parse(f_info[2][:-4].replace('_', 'T')))
        else:
            date = round_date(parse(f_info[2][:-1]))
        shutil.move(f, os.path.join(download_dir, str(channel), date.isoformat('T', timespec='hours') + '.fits'))
    [os.remove(f) for f in glob.glob(os.path.join(download_dir, '*.*'))]


def download_month(year, month):
    tstart = datetime(year, month, 1, 0, 0, 0)
    tend = tstart + relativedelta(months=1) - timedelta(hours=6)
    download_date_range(tstart, tend)


def download_date_range(tstart, tend):
    tstart, tend = tstart.isoformat('_', timespec='seconds'), tend.isoformat('_', timespec='seconds')
    print('Download AIA: %s -- %s' % (tstart, tend))
    download('aia.lev1_euv_12s[%sZ-%sZ@6h][%s]{image}' % (tstart, tend, ','.join(euv_channels)), )
    if '6173' in channels:
        print('Download HMI: %s -- %s' % (tstart, tend))
        download('hmi.M_720s[%sZ-%sZ@6h]{magnetogram}' % (tstart, tend), )


tstart = datetime(2010, 5, 1)
tend = datetime.now()
td = timedelta(days=30)

dates = [tstart + i * td for i in range((tend - tstart) // td)]
for d in dates:
    download_date_range(d, d + td)



import argparse
import glob
import os
from datetime import datetime

from dateutil.parser import parse
from sunpy.map import Map
from tqdm import tqdm

from itipy.download.download_sdo import SDODownloader

parser = argparse.ArgumentParser(description='Download SDO data aligned with SOHO observations')
parser.add_argument('--download_dir', type=str, help='path to the download directory.')
parser.add_argument('--soho_path', type=str, help='path to the reference soho data.')
parser.add_argument('--email', type=str, help='registered email address for JSOC.')

args = parser.parse_args()

downloader = SDODownloader(base_path=args.download_dir, email=args.email)
basenames_soho = [
    [os.path.basename(f) for f in glob.glob('%s/%s/*.fits' % (args.soho_path, wl))]
    for wl in ['171', '195', '284', '304', 'mag']]
basenames_soho = set(basenames_soho[0]).intersection(*basenames_soho[1:])
basenames_soho = [f for f in basenames_soho if parse(f.split('.')[0]) > datetime(2010, 5, 12)]  # coarse filter
dates = sorted(
    [Map('%s/mag/%s' % (args.soho_path, f)).date.datetime for f in basenames_soho])
dates = [d for d in dates if d > datetime(2010, 5, 12)]
for d in tqdm(dates):
    try:
        downloader.downloadDate(d)
    except Exception as ex:
        print(ex)



from warnings import simplefilter

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.nddata import block_reduce
from skimage.util import view_as_windows
from sunpy.map import Map


def alignMaps(align_map, ref_map):
    simplefilter('ignore')
    original_map = align_map
    # downscale to reference map
    scale_factor = align_map.scale[0] / ref_map.scale[0]
    align_map = Map(align_map.data, align_map.meta.copy())
    align_map = align_map.resample([int(align_map.data.shape[1] * scale_factor),
                                    int(align_map.data.shape[0] * scale_factor)] * u.pix)
    # adjust scale
    #
    width = align_map.bottom_left_coord.Tx - align_map.top_right_coord.Tx
    height = align_map.bottom_left_coord.Ty - align_map.top_right_coord.Ty
    #
    for fov, reduction_scale in zip([1.5, 1, 0.7, 0.51], [8, 4, 2, 1]):
        # initial align
        coord = align_map.center
        bl = SkyCoord(coord.Tx - width * fov, coord.Ty - height * fov, frame=ref_map.coordinate_frame)
        tr = SkyCoord(coord.Tx + width * fov, coord.Ty + height * fov, frame=ref_map.coordinate_frame)
        submap = ref_map.submap(bottom_left=bl, top_right=tr)
        #
        shift = getShift(align_map.data.astype(np.float32), submap.data.astype(np.float32),
                         reduction_block=(reduction_scale, reduction_scale))
        # apply shift
        align_map.meta['crpix1'] += shift[1]
        align_map.meta['crpix2'] += shift[0]
    # copy shift information to original image size
    new_meta = original_map.meta.copy()
    lon, lat = original_map._get_lon_lat(align_map.center.frame)
    new_meta['crpix1'] = (original_map.dimensions[0].value + 1) / 2.
    new_meta['crpix2'] = (original_map.dimensions[1].value + 1) / 2.
    new_meta['crval1'] = lon.value
    new_meta['crval2'] = lat.value
    #
    return Map(original_map.data, new_meta)


def getShift(image, image_ref, reduction_block=(1, 1)):
    """
    Pixel shift between two images by cross-correlation of the subframes.
    :param image: smaller image patch to align with the reference image
    :param image_ref: reference image
    :return: pixel shift between the image to the reference image
    """
    image = block_reduce(image, reduction_block, func=np.mean)
    image_ref = block_reduce(image_ref, reduction_block, func=np.mean)
    windows = view_as_windows(image_ref, image.shape, 1)
    #
    shifts = []
    for i in range(windows.shape[0]):
        for j in range(windows.shape[1]):
            shifts += [(i, j, correlation_coefficient(windows[i, j], image))]
    shifts = np.array(shifts)
    best_shift = shifts[np.argmax(shifts[:, 2])]
    #
    center = (image_ref.shape[0] // 2 - image.shape[0] // 2, image_ref.shape[1] // 2 - image.shape[1] // 2)
    return (center[0] - best_shift[0]) * reduction_block[0], (center[1] - best_shift[1]) * reduction_block[1]


def correlation_coefficient(patch1, patch2):
    product = np.nanmean((patch1 - np.nanmean(patch1)) * (patch2 - np.nanmean(patch2)))
    stds = np.nanstd(patch1) * np.nanstd(patch2)
    if stds == 0:
        return 0
    else:
        product /= stds
        return product



import argparse
import logging
import os
from random import sample

from itipy.data.editor import RandomPatchEditor

from itipy.data.dataset import StorageDataset, HMIContinuumDataset, HinodeDataset
from itipy.trainer import Trainer

import pandas as pd

parser = argparse.ArgumentParser(description='Train HMI-To-Hinode translations')
parser.add_argument('--base_dir', type=str, help='path to the results directory.')

parser.add_argument('--hinode_path', type=str, help='path to the Hinode data.')
parser.add_argument('--hinode_file_list', type=str, help='path to the Hinode file list (see pyiti.data.hinode.classify).')
parser.add_argument('--hmi_path', type=str, help='path to the HMI data.')
parser.add_argument('--hinode_converted_path', type=str, help='path to store the converted Hinode data.')
parser.add_argument('--hmi_converted_path', type=str, help='path to store the converted HMI data.')

args = parser.parse_args()

base_dir = args.base_dir
hmi_path = args.hmi_path
hmi_converted_path = args.hmi_converted_path
hinode_converted_path = args.hinode_converted_path
hinode_file_list = args.hinode_file_list

prediction_dir = os.path.join(base_dir, 'prediction')
os.makedirs(prediction_dir, exist_ok=True)

logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler("{0}/{1}.log".format(base_dir, "info_log")),
        logging.StreamHandler()
    ])

# Init Model
trainer = Trainer(1, 1, upsampling=2, norm='in_rs_aff', lambda_diversity=0)
trainer.cuda()

test_months = [11, 12]
train_months = list(range(2, 10))

df = pd.read_csv(hinode_file_list, index_col=False, parse_dates=['date'])

train_df = df[df.date.dt.month.isin(train_months)]
features = train_df[train_df.classification == 'feature']
quiet = train_df[train_df.classification == 'quiet']
limb = train_df[train_df.classification == 'limb']
hinode_train_files = list(features.file) + list(limb.file) + sample(list(quiet.file), len(features) + len(limb))

test_df = df[df.date.dt.month.isin(test_months)]
features = train_df[train_df.classification == 'feature']
quiet = train_df[train_df.classification == 'quiet']
limb = train_df[train_df.classification == 'limb']
hinode_test_files = list(features.file) + list(limb.file) + sample(list(quiet.file), len(features) + len(limb))

# Init Dataset
hmi_train = HMIContinuumDataset(hmi_path, (512, 512), months=train_months, ext='.fits')
hmi_train = StorageDataset(hmi_train, hmi_converted_path, ext_editors=[RandomPatchEditor((160, 160))])

hmi_valid = HMIContinuumDataset(hmi_path, (512, 512), months=test_months, ext='.fits')
hmi_valid = StorageDataset(hmi_valid, hmi_converted_path, ext_editors=[RandomPatchEditor((160, 160))])

hinode_train = HinodeDataset(hinode_train_files)
hinode_train = StorageDataset(hinode_train, hinode_converted_path,
                              ext_editors=[RandomPatchEditor((640, 640))])

hinode_valid = HinodeDataset(hinode_test_files)
hinode_valid = StorageDataset(hinode_valid, hinode_converted_path, ext_editors=[RandomPatchEditor((640, 640))])

plot_settings_A = [
    {"cmap": "gray", "title": "HMI Continuum", 'vmin': -1, 'vmax': 1, },
]
plot_settings_B = [
    {"cmap": "gray", "title": "Hinode Continuum", 'vmin': -1, 'vmax': 1},
]

trainer.startBasicTraining(base_dir, hmi_train, hinode_train, hmi_valid, hinode_valid, plot_settings_A, plot_settings_B)



import os
import os
import random
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from random import randint
from urllib import request

import astropy.io.ascii
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from aiapy.calibrate import correct_degradation
from aiapy.calibrate.util import get_correction_table
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.visualization import ImageNormalize, LinearStretch, AsinhStretch
from dateutil.parser import parse
from scipy import ndimage
from skimage.measure import block_reduce
from skimage.transform import pyramid_reduce
from sunpy.coordinates import frames
from sunpy.map import Map, all_coordinates_from_map, header_helper


class Editor(ABC):

    def convert(self, data, **kwargs):
        result = self.call(data, **kwargs)
        if isinstance(result, tuple):
            data, add_kwargs = result
            kwargs.update(add_kwargs)
        else:
            data = result
        return data, kwargs

    @abstractmethod
    def call(self, data, **kwargs):
        raise NotImplementedError()


sdo_norms = {94: ImageNormalize(vmin=0, vmax=340, stretch=AsinhStretch(0.005), clip=True),
             131: ImageNormalize(vmin=0, vmax=1400, stretch=AsinhStretch(0.005), clip=True),
             171: ImageNormalize(vmin=0, vmax=8600, stretch=AsinhStretch(0.005), clip=True),
             193: ImageNormalize(vmin=0, vmax=9800, stretch=AsinhStretch(0.005), clip=True),
             211: ImageNormalize(vmin=0, vmax=5800, stretch=AsinhStretch(0.005), clip=True),
             304: ImageNormalize(vmin=0, vmax=8800, stretch=AsinhStretch(0.001), clip=True),
             335: ImageNormalize(vmin=0, vmax=600, stretch=AsinhStretch(0.005), clip=True),
             1600: ImageNormalize(vmin=0, vmax=4000, stretch=AsinhStretch(0.005), clip=True),
             1700: ImageNormalize(vmin=0, vmax=4000, stretch=AsinhStretch(0.005), clip=True),
             'mag': ImageNormalize(vmin=-3000, vmax=3000, stretch=LinearStretch(), clip=True),
             'continuum': ImageNormalize(vmin=0, vmax=70000, stretch=LinearStretch(), clip=True),
             }

soho_norms = {171: ImageNormalize(vmin=0, vmax=16000, stretch=AsinhStretch(0.005), clip=True),
              195: ImageNormalize(vmin=0, vmax=12000, stretch=AsinhStretch(0.005), clip=True),
              284: ImageNormalize(vmin=0, vmax=2300, stretch=AsinhStretch(0.005), clip=True),
              304: ImageNormalize(vmin=0, vmax=11000, stretch=AsinhStretch(0.005), clip=True),
              6173: ImageNormalize(vmin=-3000, vmax=3000, stretch=LinearStretch(), clip=True),
              }

stereo_norms = {171: ImageNormalize(vmin=0, vmax=6000, stretch=AsinhStretch(0.005), clip=True),
                195: ImageNormalize(vmin=0, vmax=3400, stretch=AsinhStretch(0.005), clip=True),
                284: ImageNormalize(vmin=0, vmax=1300, stretch=AsinhStretch(0.005), clip=True),
                304: ImageNormalize(vmin=0, vmax=18100, stretch=AsinhStretch(0.005), clip=True),
                }

hinode_norms = {'continuum': ImageNormalize(vmin=0, vmax=50000, stretch=LinearStretch(), clip=True),
                'gband': ImageNormalize(vmin=0, vmax=25000, stretch=LinearStretch(), clip=True), }

gregor_norms = {'gband': ImageNormalize(vmin=0, vmax=1.8, stretch=LinearStretch(), clip=True)}


class LoadFITSEditor(Editor):

    def call(self, map_path, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # ignore warnings
            hdul = fits.open(map_path)
            hdul.verify("fix")
            data, header = hdul[0].data, hdul[0].header
            hdul.close()
        return data, {"header": header}


class LoadMapEditor(Editor):

    def call(self, data, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            s_map = Map(data)
            s_map.meta['timesys'] = 'tai'  # fix leap seconds
            return s_map, {'path': data}


class LoadGregorGBandEditor(Editor):

    def call(self, file, **kwargs):
        warnings.simplefilter("ignore")
        hdul = fits.open(file)
        #
        assert 'wavelnth' in hdul[0].header, 'Invalid GREGOR file %s' % file
        if hdul[0].header['wavelnth'] == 430.7:
            index = 0
        elif hdul[1].header['wavelnth'] == 430.7:
            index = 1
        else:
            raise Exception('Invalid GREGOR file %s' % file)
        #
        primary_header = hdul[0].header
        primary_header['cunit1'] = 'arcsec'
        primary_header['cunit2'] = 'arcsec'
        primary_header['cdelt1'] = 70 / 1280
        primary_header['cdelt2'] = 70 / 1280
        #
        g_band = hdul[index::2]
        g_band = sorted(g_band, key=lambda hdu: hdu.header['TIMEOFFS'])
        #
        gregor_maps = [Map(hdu.data, primary_header) for hdu in g_band]
        return gregor_maps, {'path': file}


class SubMapEditor(Editor):

    def __init__(self, coords):
        self.coords = coords

    def call(self, s_map, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # ignore warnings
            return s_map.submap(SkyCoord(*self.coords, frame=s_map.coordinate_frame))


class MapToDataEditor(Editor):

    def call(self, s_map, **kwargs):
        return s_map.data, {"header": s_map.meta}


class AddRadialDistanceEditor(Editor):

    def call(self, data, **kwargs):
        s_map = Map(data, kwargs['header'])
        coords = all_coordinates_from_map(s_map)
        radial_distance = (np.sqrt(coords.Tx ** 2 + coords.Ty ** 2) / s_map.rsun_obs).value
        radial_distance[radial_distance >= 1] = -1
        return np.stack([s_map.data, radial_distance])


class DataToMapEditor(Editor):

    def call(self, data, **kwargs):
        return Map(data[0], kwargs['header'])


class ContrastNormalizeEditor(Editor):

    def __init__(self, use_median=False, shift=None, normalization=None):
        self.use_median = use_median
        self.shift = shift
        self.normalization = normalization

    def call(self, data, **kwargs):
        if self.shift is None:
            shift = np.nanmedian(data) if self.use_median else np.nanmean(data)
        else:
            shift = self.shift
        if self.normalization is None:
            normalization = np.nanstd(data)
        else:
            normalization = self.normalization
        data = (data - shift) / (normalization + 10e-8)
        return data


class ImageNormalizeEditor(Editor):

    def __init__(self, vmin=None, vmax=None, stretch=LinearStretch()):
        self.norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=stretch, clip=True)

    def call(self, data, **kwargs):
        data = self.norm(data).data * 2 - 1
        return data


class NormalizeEditor(Editor):

    def __init__(self, norm):
        self.norm = norm

    def call(self, data, **kwargs):
        data = self.norm(data).data * 2 - 1
        return data


class ReshapeEditor(Editor):

    def __init__(self, shape):
        self.shape = shape

    def call(self, data, **kwargs):
        data = data[:self.shape[1], :self.shape[2]]
        return np.reshape(data, self.shape).astype(np.float32)


class ExpandDimsEditor(Editor):

    def __init__(self, axis=0):
        self.axis = axis

    def call(self, data, **kwargs):
        return np.expand_dims(data, axis=self.axis).astype(np.float32)


class NanEditor(Editor):
    def __init__(self, nan=0):
        self.nan = nan

    def call(self, data, **kwargs):
        data = np.nan_to_num(data, nan=self.nan)
        return data


class KSOPrepEditor(Editor):
    def __init__(self, add_rotation=False):
        self.add_rotation = add_rotation

        super().__init__()

    def call(self, kso_map, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # ignore warnings
            kso_map.meta["waveunit"] = "AA"
            kso_map.meta["arcs_pp"] = kso_map.scale[0].value
            if 'exptime' not in kso_map.meta and 'exp_time' in kso_map.meta:
                kso_map.meta['exptime'] = kso_map.meta['exp_time'] / 1000  # ms to s

            if self.add_rotation:
                angle = -kso_map.meta["angle"]
            else:
                angle = 0
            c = np.cos(np.deg2rad(angle))
            s = np.sin(np.deg2rad(angle))

            kso_map.meta["PC1_1"] = c
            kso_map.meta["PC1_2"] = -s
            kso_map.meta["PC2_1"] = s
            kso_map.meta["PC2_2"] = c

            return kso_map


class KSOFilmPrepEditor(Editor):
    def __init__(self, add_rotation=False):
        self.add_rotation = add_rotation

        super().__init__()

    def call(self, data, **kwargs):
        data = data[0]
        h = kwargs['header']
        coord = SkyCoord(0 * u.arcsec, 0 * u.arcsec, obstime=parse(h['DATE_OBS']), observer='earth',
                         frame=frames.Helioprojective)
        header = header_helper.make_fitswcs_header(
            data, coord,
            rotation_angle=h["ANGLE"] * u.deg if self.add_rotation else 0 * u.deg,
            reference_pixel=u.Quantity([h['CENTER_X'], h['CENTER_Y']] * u.pixel),
            scale=u.Quantity([h['CDELT1'], h['CDELT2']] * u.arcsec / u.pix),
            instrument=h["INSTRUME"],
            exposure=h["EXPTIME"] * u.ms,
            wavelength=h["WAVELNTH"] * u.angstrom)

        return Map(data, header)


class AIAPrepEditor(Editor):
    def __init__(self, calibration='auto'):
        super().__init__()
        assert calibration in ['aiapy', 'auto', 'none',
                               None], "Calibration must be one of: ['aiapy', 'auto', 'none', None]"
        self.calibration = calibration
        self.table = get_auto_calibration_table() if calibration == 'auto' else get_local_correction_table()

    def call(self, s_map, **kwargs):
        warnings.simplefilter("ignore")  # ignore warnings
        if self.calibration == 'auto':
            s_map = self.correct_degradation(s_map, correction_table=self.table)
        elif self.calibration == 'aiapy':
            s_map = correct_degradation(s_map, correction_table=self.table)
        data = np.nan_to_num(s_map.data)
        data = data / s_map.meta["exptime"]
        return Map(data.astype(np.float32), s_map.meta)

    def correct_degradation(self, s_map, correction_table):
        index = correction_table["DATE"].sub(s_map.date.datetime).abs().idxmin()
        num = s_map.meta["wavelnth"]
        return Map(s_map.data / correction_table.iloc[index][f"{int(num):04}"], s_map.meta)


class NormalizeExposureEditor(Editor):
    def __init__(self, target=1 * u.s):
        self.target = target
        super().__init__()

    def call(self, s_map, **kwargs):
        warnings.simplefilter("ignore")  # ignore warnings
        data = s_map.data
        data = data / s_map.exposure_time.to(u.s).value * self.target.to(u.s).value
        return Map(data.astype(np.float32), s_map.meta)


class NormalizeRadiusEditor(Editor):
    def __init__(self, resolution, padding_factor=0.1, crop=True, fix_irradiance_with_distance=False, **kwargs):
        self.padding_factor = padding_factor
        self.resolution = resolution
        self.crop = crop
        self.fix_irradiance_with_distance = fix_irradiance_with_distance
        super(NormalizeRadiusEditor, self).__init__(**kwargs)

    def call(self, s_map, **kwargs):
        warnings.simplefilter("ignore")  # ignore warnings

        if self.fix_irradiance_with_distance:
            old_meta = s_map.meta.copy()
        r_obs_pix = s_map.rsun_obs / s_map.scale[0]  # Get the solar radius in pixels
        r_obs_pix = (1 + self.padding_factor) * r_obs_pix  # Get the size in pixels of the padded radius 
        scale_factor = self.resolution / (2 * r_obs_pix.value)
        s_map = Map(np.nan_to_num(s_map.data).astype(np.float32), s_map.meta)
        s_map = s_map.rotate(recenter=True, scale=scale_factor, missing=0, order=4)
        if self.crop:
            arcs_frame = (self.resolution / 2) * s_map.scale[0].value
            s_map = s_map.submap(bottom_left=SkyCoord(-arcs_frame * u.arcsec, -arcs_frame * u.arcsec, frame=s_map.coordinate_frame),
                                 top_right=SkyCoord(arcs_frame * u.arcsec, arcs_frame * u.arcsec, frame=s_map.coordinate_frame))
            pad_x = s_map.data.shape[0] - self.resolution
            pad_y = s_map.data.shape[1] - self.resolution
            s_map = s_map.submap(bottom_left=[pad_x // 2, pad_y // 2] * u.pix,
                                 top_right=[pad_x // 2 + self.resolution - 1, pad_y // 2 + self.resolution - 1] * u.pix)
        
        s_map.meta['r_sun'] = s_map.rsun_obs.value / s_map.meta['cdelt1']

        # Virtually move the instrument such that the sun occupies the expected
        # size in the current optics
        if self.fix_irradiance_with_distance:
            # The sun is bigger, not the scaling of the detector
            s_map.meta['rsun_obs'] = old_meta['rsun_obs']*scale_factor
            # This means that cdelt is the same as the old one
            s_map.meta['cdelt1'] = old_meta['cdelt1']
            s_map.meta['cdelt2'] = old_meta['cdelt1']           
            # But we are also closer to the sun
            s_map.meta['dsun_obs'] = (s_map.meta['rsun_ref']/np.tan(s_map.meta['rsun_obs']*u.arcsec)).value
            # Change intensity due to distance change
            s_map.data[:] = s_map.data[:] * (old_meta['dsun_obs']**2)/(s_map.meta['dsun_obs']**2)

        return s_map


class RecenterEditor(Editor):

    def __init__(self, missing=0, order=4, **kwargs):
        self.missing = missing
        self.order = order
        super().__init__(**kwargs)

    def call(self, s_map, **kwargs):
        return s_map.rotate(recenter=True, missing=self.missing, order=self.order)


class ScaleEditor(Editor):
    def __init__(self, arcspp):
        self.arcspp = arcspp
        super(ScaleEditor, self).__init__()

    def call(self, s_map, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # ignore warnings
            scale_factor = s_map.scale[0].value / self.arcspp
            new_dimensions = [int(s_map.data.shape[1] * scale_factor),
                              int(s_map.data.shape[0] * scale_factor)] * u.pixel
            s_map = s_map.resample(new_dimensions)
            return Map(s_map.data.astype(np.float32), s_map.meta)


class SubmapSolarRadiiEditor(Editor):
    def __init__(self, solar_radii=1):
        self.solar_radii = solar_radii
        super(SubmapSolarRadiiEditor, self).__init__()

    def call(self, s_map, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # ignore warnings
            arcs_frame = s_map.rsun_obs * self.solar_radii
            s_map = s_map.submap(SkyCoord([-arcs_frame, arcs_frame] * u.arcsec,
                                          [-arcs_frame, arcs_frame] * u.arcsec,
                                          frame=s_map.coordinate_frame))

            return s_map


class PyramidRescaleEditor(Editor):

    def __init__(self, scale=2):
        self.scale = scale

    def call(self, data, **kwargs):
        if self.scale == 1:
            return data
        data = pyramid_reduce(data, downscale=self.scale)
        return data


class BlockReduceEditor(Editor):

    def __init__(self, block_size, func=np.mean):
        self.block_size = block_size
        self.func = func

    def call(self, data, **kwargs):
        return block_reduce(data, self.block_size, func=self.func)


class LoadNumpyEditor(Editor):

    def call(self, data, **kwargs):
        return np.load(data)


class StackEditor(Editor):

    def __init__(self, data_sets):
        self.data_sets = data_sets

    def call(self, idx, **kwargs):
        results = [dp.getIndex(idx) for dp in self.data_sets]
        return np.concatenate([img for img, kwargs in results], 0), {'kwargs_list': [kwargs for img, kwargs in results]}


class DistributeEditor(Editor):

    def __init__(self, editors):
        self.editors = editors

    def call(self, data, **kwargs):
        return np.concatenate([self.convertData(d, **kwargs) for d in data], 0)

    def convertData(self, data, **kwargs):
        for editor in self.editors:
            data, kwargs = editor.convert(data, **kwargs)
        return data


class RemoveOffLimbEditor(Editor):

    def __init__(self, fill_value=0):
        self.fill_value = fill_value

    def call(self, s_map, **kwargs):
        warnings.simplefilter("ignore")  # ignore warnings
        hpc_coords = all_coordinates_from_map(s_map)
        r = np.sqrt(hpc_coords.Tx ** 2 + hpc_coords.Ty ** 2) / s_map.rsun_obs
        s_map.data[r > 1] = self.fill_value
        return s_map


class FeaturePatchEditor(Editor):

    def __init__(self, patch_shape=(512, 512)):
        self.patch_shape = patch_shape

    def call(self, s_map, **kwargs):
        warnings.simplefilter("ignore")  # ignore warnings
        hpc_coords = all_coordinates_from_map(s_map)
        r = np.sqrt(hpc_coords.Tx ** 2 + hpc_coords.Ty ** 2) / s_map.rsun_obs

        data = np.copy(s_map.data)
        # randomly select limb
        if random.random() < 0.2:
            data[r > 0.9] = np.nan
            data[(r > 0.9) & (r < 1.)] = -1
        else:
            data[r > 0.9] = np.nan

        pixel_pos = np.argwhere(data == np.nanmin(data))
        pixel_pos = pixel_pos[randint(0, len(pixel_pos) - 1)]
        pixel_pos = np.min([pixel_pos[0], data.shape[0] - self.patch_shape[0] // 2]), \
                    np.min([pixel_pos[1], data.shape[1] - self.patch_shape[1] // 2])
        pixel_pos = np.max([pixel_pos[0], self.patch_shape[0] // 2]), \
                    np.max([pixel_pos[1], self.patch_shape[1] // 2])

        pixel_pos = np.array(pixel_pos) * u.pix
        center = s_map.pixel_to_world(pixel_pos[1], pixel_pos[0])

        arcs_frame = s_map.scale[0] * (self.patch_shape[0] / 2 * u.pix)
        s_map = s_map.submap(SkyCoord([center.Tx - arcs_frame, center.Tx + arcs_frame],
                                      [center.Ty - arcs_frame, center.Ty + arcs_frame],
                                      frame=s_map.coordinate_frame))

        return s_map


class RandomPatchEditor(Editor):
    def __init__(self, patch_shape):
        self.patch_shape = patch_shape

    def call(self, data, **kwargs):
        assert data.shape[1] >= self.patch_shape[0], 'Invalid data shape: %s' % str(data.shape)
        assert data.shape[2] >= self.patch_shape[1], 'Invalid data shape: %s' % str(data.shape)
        x = randint(0, data.shape[1] - self.patch_shape[0])
        y = randint(0, data.shape[2] - self.patch_shape[1])
        patch = data[:, x:x + self.patch_shape[0], y:y + self.patch_shape[1]]
        patch = np.copy(patch)  # copy from mmep
        assert np.std(patch) != 0, 'Invalid patch found (all values %f)' % np.mean(patch)
        assert not np.any(np.isnan(patch)), 'NaN found'
        return patch


class RandomPatch3DEditor(Editor):
    def __init__(self, patch_shape):
        self.patch_shape = patch_shape

    def call(self, data, **kwargs):
        assert data.shape[0] >= self.patch_shape[0], 'Invalid data shape: %s' % str(data.shape)
        assert data.shape[1] >= self.patch_shape[1], 'Invalid data shape: %s' % str(data.shape)
        assert data.shape[2] >= self.patch_shape[2], 'Invalid data shape: %s' % str(data.shape)
        c = randint(0, data.shape[0] - self.patch_shape[0])
        x = randint(0, data.shape[1] - self.patch_shape[1])
        y = randint(0, data.shape[2] - self.patch_shape[2])
        patch = data[c:c + self.patch_shape[0], x:x + self.patch_shape[1], y:y + self.patch_shape[2]]
        patch = np.copy(patch)  # copy from mmep
        assert np.std(patch) != 0, 'Invalid patch found (all values %f)' % np.mean(patch)
        assert not np.any(np.isnan(patch)), 'NaN found'
        return patch


class SliceEditor(Editor):

    def __init__(self, start, stop):
        self.start = start
        self.stop = stop

    def call(self, data, **kwargs):
        return data[self.start:self.stop]


class BrightestPixelPatchEditor(Editor):
    def __init__(self, patch_shape, idx=0, random_selection=0.2):
        self.patch_shape = patch_shape
        self.idx = idx
        self.random_selection = random_selection

    def call(self, data, **kwargs):
        assert data.shape[1] >= self.patch_shape[0], 'Invalid data shape: %s' % str(data.shape)
        assert data.shape[2] >= self.patch_shape[1], 'Invalid data shape: %s' % str(data.shape)

        if random.random() <= self.random_selection:
            x = randint(0, data.shape[1] - self.patch_shape[0])
            y = randint(0, data.shape[2] - self.patch_shape[1])
            patch = data[:, x:x + self.patch_shape[0], y:y + self.patch_shape[1]]
        else:
            smoothed = ndimage.gaussian_filter(data[self.idx], sigma=5)
            pixel_pos = np.argwhere(smoothed == np.nanmax(smoothed))
            pixel_pos = pixel_pos[randint(0, len(pixel_pos) - 1)]
            pixel_pos = np.min([pixel_pos[0], smoothed.shape[0] - self.patch_shape[0] // 2]), np.min(
                [pixel_pos[1], smoothed.shape[1] - self.patch_shape[1] // 2])
            pixel_pos = np.max([pixel_pos[0], self.patch_shape[0] // 2]), np.max(
                [pixel_pos[1], self.patch_shape[1] // 2])

            x = pixel_pos[0]
            y = pixel_pos[1]
            patch = data[:,
                    x - int(np.floor(self.patch_shape[0] / 2)):x + int(np.ceil(self.patch_shape[0] / 2)),
                    y - int(np.floor(self.patch_shape[1] / 2)):y + int(np.ceil(self.patch_shape[1] / 2)), ]
        assert np.std(patch) != 0, 'Invalid patch found (all values %f)' % np.mean(patch)
        return patch


class EITCheckEditor(Editor):

    def call(self, s_map, **kwargs):
        assert np.all(np.logical_not(np.isnan(s_map.data))), 'Found missing block %s' % s_map.date.datetime.isoformat()
        assert 'N_MISSING_BLOCKS =    0' in s_map.meta['comment'], 'Found missing block %s: %s' % (
            s_map.date.datetime.isoformat(), s_map.meta['comment'])
        return s_map

class SOHOFixHeaderEditor(Editor):

    def call(self, s_map, **kwargs):
        s_map.meta['DATE-OBS'] = s_map.meta['DATE_OBS']  # fix date
        s_map.meta['rsun_ref'] = s_map.rsun_meters.value  # preserve solar radius (SOHO fix)
        return s_map

class SECCHIPrepEditor(Editor):

    def __init__(self, degradation=None):
        self.degradation_fit = np.poly1d(degradation) if degradation else False

    def call(self, s_map, **kwargs):
        assert np.all(np.logical_not(np.isnan(s_map.data))), 'Found missing block %s' % s_map.date.datetime.isoformat()
        assert s_map.meta['nmissing'] == 0, 'Found missing block %s: %s' % (
        s_map.date.datetime.isoformat(), s_map.meta['nmissing'])
        assert s_map.meta['NAXIS1'] == 2048 and s_map.meta[
            'NAXIS2'] == 2048, 'Found invalid resolution: %s' % s_map.date.datetime.isoformat()
        if self.degradation_fit:
            x = mdates.date2num(s_map.date.datetime)
            correction = self.degradation_fit(x)
            s_map = Map(s_map.data / correction, s_map.meta)
        return s_map


class PaddingEditor(Editor):
    def __init__(self, target_shape):
        self.target_shape = target_shape

    def call(self, data, **kwargs):
        s = data.shape
        p = self.target_shape
        x_pad = (p[0] - s[-2]) / 2
        y_pad = (p[1] - s[-1]) / 2
        pad = [(int(np.floor(x_pad)), int(np.ceil(x_pad))),
               (int(np.floor(y_pad)), int(np.ceil(y_pad)))]
        if len(s) == 3:
            pad.insert(0, (0, 0))
        return np.pad(data, pad, 'constant', constant_values=np.nan)


class UnpaddingEditor(Editor):
    def __init__(self, target_shape):
        self.target_shape = target_shape

    def call(self, data, **kwargs):
        s = data.shape
        p = self.target_shape
        x_unpad = (s[-2] - p[0]) / 2
        y_unpad = (s[-1] - p[1]) / 2
        #
        unpad = [(None if int(np.floor(y_unpad)) == 0 else int(np.floor(y_unpad)),
                  None if int(np.ceil(y_unpad)) == 0 else -int(np.ceil(y_unpad))),
                 (None if int(np.floor(x_unpad)) == 0 else int(np.floor(x_unpad)),
                  None if int(np.ceil(x_unpad)) == 0 else -int(np.ceil(x_unpad)))]
        data = data[:, unpad[0][0]:unpad[0][1], unpad[1][0]:unpad[1][1]]
        return data


class ReductionEditor(Editor):

    def call(self, data, **kwargs):
        s = data.shape
        p = kwargs['patch_shape']
        x_pad = (s[-2] - p[-2]) / 2
        y_pad = (s[-1] - p[-1]) / 2
        pad = [(int(np.floor(x_pad)), int(np.ceil(x_pad))),
               (int(np.floor(y_pad)), int(np.ceil(y_pad)))]
        if x_pad == 0 and y_pad == 0:
            return data
        if x_pad == 0:
            return data[..., pad[1][0]:-pad[1][1]]
        if y_pad == 0:
            return data[..., pad[0][0]:-pad[0][1], :]
        return data[..., pad[0][0]:-pad[0][1], pad[1][0]:-pad[1][1]]


class PassEditor(Editor):

    def call(self, data, **kwargs):
        return data


class LambdaEditor(Editor):

    def __init__(self, f):
        self.f = f

    def call(self, data, **kwargs):
        return self.f(data, **kwargs)


class LimbDarkeningCorrectionEditor(Editor):

    def __init__(self, limb_offset=0.99):
        self.limb_offset = limb_offset

    def call(self, s_map, **kwargs):
        coords = all_coordinates_from_map(s_map)
        radial_distance = (np.sqrt(coords.Tx ** 2 + coords.Ty ** 2) / s_map.rsun_obs).value
        radial_distance[radial_distance >= self.limb_offset] = np.NaN
        ideal_correction = np.cos(radial_distance * np.pi / 2)

        condition = np.logical_not(np.isnan(np.ravel(ideal_correction)))
        map_list = np.ravel(s_map.data)[condition]
        correction_list = np.ravel(ideal_correction)[condition]

        fit = np.polyfit(correction_list, map_list, 4)
        poly_fit = np.poly1d(fit)

        map_correction = poly_fit(ideal_correction)
        corrected_map = s_map.data / map_correction

        return Map(corrected_map, s_map.meta)


def get_local_correction_table():
    path = os.path.join(Path.home(), 'aiapy', 'correction_table.dat')
    if os.path.exists(path):
        return get_correction_table(path)
    os.makedirs(os.path.join(Path.home(), 'aiapy'), exist_ok=True)
    correction_table = get_correction_table()
    astropy.io.ascii.write(correction_table, path)
    return correction_table


def get_auto_calibration_table():
    table_path = os.path.join(Path.home(), '.iti', 'sdo_autocal_table.csv')
    os.makedirs(os.path.join(Path.home(), '.iti'), exist_ok=True)
    if not os.path.exists(table_path):
        request.urlretrieve('http://kanzelhohe.uni-graz.at/iti/sdo_autocal_table.csv', filename=table_path)
    return pd.read_csv(table_path, parse_dates=['DATE'], index_col=0)



from enum import Enum

import torch
from torch import nn
from torch.autograd import Variable
from torch.nn import LayerNorm, init


class DiscriminatorMode(Enum):
    SINGLE = "SINGLE"  # use a single discriminator across all channels
    CHANNELS = "CHANNELS"  # use a discriminator per channel and one for the the combined channels
    SINGLE_PER_CHANNEL = "SINGLE_PER_CHANNEL"  # use a single discriminator for each channel and one for the combined channels


########################## Generator ##################################
class GeneratorAB(nn.Module):
    def __init__(self, input_dim, output_dim, depth, n_upsample, dim=64, output_activ='tanh', skip_connections=True, **kwargs):
        super().__init__()
        self.depth = depth
        # self.skip_connections = skip_connections
        self.from_image = Conv2dBlock(input_dim, dim, 7, 1, 3, **kwargs)
        ##################### Encoder #####################
        self.down_blocks = []
        n_convs = 1
        for i in range(depth):
            self.down_blocks += [DownBlock(dim, 2 * dim, n_convs, **kwargs)]
            dim *= 2
            n_convs = n_convs + 1 if n_convs < 3 else 3
        ##################### Core #####################
        self.core_block = CoreBlock(dim, dim, 3, **kwargs)
        ##################### Decoder #####################
        self.up_blocks = []
        for i in range(depth):
            self.up_blocks += [UpBlock(dim, dim // 2, n_convs, skip_connection=skip_connections, **kwargs)]
            dim //= 2
            n_convs = n_convs -1 if n_convs > 1 else 1
        ##################### Upsampling #####################
        self.sampling_blocks = []
        for _ in range(n_upsample):
            self.sampling_blocks += [UpBlock(dim, dim // 2, 1, skip_connection=False, **kwargs)]
            dim //= 2
        self.to_image = Conv2dBlock(dim, output_dim, 7, 1, 3, norm='none', activation=output_activ, pad_type=kwargs['pad_type'])

        self.model = nn.Sequential(self.from_image, *self.down_blocks, self.core_block, *self.up_blocks, *self.sampling_blocks, self.to_image)

    def forward(self, x):
        self.skip_connections = True
        x = self.from_image(x)
        # encode
        skip_connections = []
        for down in self.down_blocks:
            x, skip = down(x)
            skip_connections.append(skip)
        # core
        x = self.core_block(x)
        # decode
        for up, skip in zip(self.up_blocks, reversed(skip_connections)):
            if self.skip_connections:
                x = up(x, skip)
            else:
                x = up(x)
        # upsampling
        for up in self.sampling_blocks:
            x = up(x)
        x = self.to_image(x)
        return x

class GeneratorBA(nn.Module):
    def __init__(self, input_dim, output_dim, noise_dim, depth, depth_noise, n_downsample, dim=64, output_activ='tanh', skip_connections=True, **kwargs):
        super().__init__()
        self.depth = depth
        self.noise_dim = noise_dim
        self.depth_noise = depth_noise
        self.n_downsample = n_downsample
        self.skip_connections = skip_connections
        dim = dim // (2 ** n_downsample)
        self.from_image = Conv2dBlock(input_dim, dim, 7, 1, 3, **kwargs)
        ##################### Downsampling #####################
        self.sampling_blocks = []
        for _ in range(n_downsample):
            self.sampling_blocks += [DownBlock(dim, dim * 2, 1, **kwargs)]
            dim *= 2
        ##################### Encoder #####################
        self.down_blocks = []
        n_convs = 1
        for i in range(depth):
            self.down_blocks += [DownBlock(dim, 2 * dim, n_convs, **kwargs)]
            dim *= 2
            n_convs = n_convs + 1 if n_convs < 3 else 3
        ##################### Noise #####################
        self.noise_blocks = []
        self.noise_blocks += [Conv2dBlock(noise_dim, dim, 7, 1, 3, **kwargs)]
        for _ in range(depth_noise - depth):
            self.noise_blocks += [UpBlock(dim, dim, 3, skip_connection=False, **kwargs)]
        self.noise_blocks = nn.Sequential(*self.noise_blocks)
        ##################### Core #####################
        self.core_block = CoreBlock(dim * 2, dim, 3, **kwargs)
        ##################### Decoder #####################
        self.up_blocks = []
        for i in range(depth):
            self.up_blocks += [UpBlock(dim, dim // 2, n_convs, skip_connection=skip_connections, **kwargs)]
            dim //= 2
            n_convs = n_convs - 1 if n_convs > 1 else 1
        self.to_image = Conv2dBlock(dim, output_dim, 7, 1, 3, norm='none', activation=output_activ, pad_type=kwargs['pad_type'])

        self.model = nn.Sequential(self.from_image, *self.sampling_blocks, *self.down_blocks, self.noise_blocks, self.core_block, *self.up_blocks, self.to_image)

    def forward(self, image, noise):
        x = self.from_image(image)
        # downsampling
        for down in self.sampling_blocks:
            x, _ = down(x)
        # encode
        skip_connections = []
        for down in self.down_blocks:
            x, skip = down(x)
            skip_connections.append(skip)
        # noise
        y = self.noise_blocks(noise)
        # core
        x = torch.cat([x, y], dim=1)
        x = self.core_block(x)
        # decode
        for up, skip in zip(self.up_blocks, reversed(skip_connections)):
            if hasattr(self, 'skip_connections') and self.skip_connections: # check for backwards compatibility
                x = up(x, skip)
            else:
                x = up(x)

        x = self.to_image(x)
        return x

    def forwardRandomNoise(self, image):
        n_gen = Variable(torch.rand(image.size(0), self.noise_dim,
                                    image.size(2) // 2 ** (self.depth_noise + self.n_downsample),
                                    image.size(3) // 2 ** (self.depth_noise + self.n_downsample)).cuda())
        return self.forward(image, noise=n_gen)


class Discriminator(nn.Module):
    def __init__(self, input_dim, n_filters, num_scales=3, depth_discriminator=4,
                 discriminator_mode=DiscriminatorMode.SINGLE,
                 norm='in_rs_aff', batch_statistic=False):
        self.pad_type = 'reflect'
        self.activ = 'relu'
        self.batch_statistic = batch_statistic
        self.norm = norm
        self.depth_discriminator = depth_discriminator
        super().__init__()
        self.input_dim = input_dim
        self.num_scales = num_scales
        self.downsample = nn.AvgPool2d(3, stride=2, padding=[1, 1], count_include_pad=False)

        self.discs = nn.ModuleList()
        self.channel_discs = nn.ModuleDict()
        # create combined discriminators
        for _ in range(num_scales):
            self.discs.append(self._make_net(input_dim, n_filters))
        # create channel discriminators
        if discriminator_mode == DiscriminatorMode.CHANNELS:
            for i in range(input_dim):
                channel_disc = nn.ModuleList()
                for _ in range(num_scales):
                    channel_disc.append(self._make_net(1, n_filters))
                self.channel_discs['%d' % i] = channel_disc
        if discriminator_mode == DiscriminatorMode.SINGLE_PER_CHANNEL:
            channel_disc = nn.ModuleList()
            for _ in range(num_scales):
                channel_disc.append(self._make_net(1, n_filters))
            for i in range(input_dim):
                self.channel_discs['%d' % i] = channel_disc

    def _make_net(self, input_dim, dim=64):
        cnn_x = []
        if self.batch_statistic:
            cnn_x += [BatchStatistic()]
        cnn_x += [Conv2dBlock((input_dim * 3) if self.batch_statistic else input_dim, dim, 4, 2, 1, norm='none', activation=self.activ, pad_type=self.pad_type)]
        for i in range(self.depth_discriminator - 1):
            cnn_x += [Conv2dBlock(dim, dim * 2, 4, 2, 1, norm=self.norm, activation=self.activ, pad_type=self.pad_type)]
            dim *= 2
        cnn_x += [nn.Conv2d(dim, 1, 1, 1, 0)]
        cnn_x = nn.Sequential(*cnn_x)
        return cnn_x

    def forward(self, x):
        outputs = []
        for i in range(self.num_scales):
            outputs.append(self.discs[i](x))
            for j, discs in enumerate(self.channel_discs.values()):
                outputs.append(discs[i](x[:, j:j + 1]))
            x = self.downsample(x)
        return outputs

    def calc_dis_loss(self, input_fake, input_real):
        # calculate the loss to train D
        outs0 = self.forward(input_fake)
        outs1 = self.forward(input_real)
        loss = 0

        for it, (out0, out1) in enumerate(zip(outs0, outs1)):
            loss += torch.mean((out0 - 0) ** 2) + torch.mean((out1 - 1) ** 2)  # LSGAN
        # normalize for Discriminators
        return loss / len(outs0)

    def calc_gen_loss(self, input_fake):
        # calculate the loss to train G
        outs0 = self.forward(input_fake)
        loss = 0
        for it, (out0) in enumerate(outs0):
            loss += torch.mean((out0 - 1) ** 2)  # LSGAN
        # normalize for Discriminators
        return loss / len(outs0)

    def calc_content_loss(self, input_real, input_fake):
        loss = []
        for i in range(self.num_scales):
            # content loss of combined discriminator
            x = input_real
            y = input_fake
            for j, layer in enumerate(self.discs[i][:-1]):
                x = layer(x)
                y = layer(y)
                if j == 0: # skip first layer (no normalization)
                    continue
                loss.append(torch.mean(torch.abs(x - y), [1, 2, 3]))

            # content loss of channel discriminator
            for j, discs in enumerate(self.channel_discs.values()):
                x = input_real[:, j:j + 1]
                y = input_fake[:, j:j + 1]
                for k, layer in enumerate(discs[i][:-1]):
                    x = layer(x)
                    y = layer(y)
                    if k == 0:  # skip first layer (no normalization)
                        continue
                    loss.append(torch.mean(torch.abs(x - y), [1, 2, 3]))

            input_real = self.downsample(input_real)
            input_fake = self.downsample(input_fake)
        return torch.mean(torch.stack(loss, 1), 1)

    def calc_content_map(self, input_real, input_fake, skip_last=1):
        loss = []
        for i in range(self.num_scales):
            # content loss of combined discriminator
            x = input_real
            y = input_fake
            for j, layer in enumerate(self.discs[i][:-skip_last]):
                up = nn.UpsamplingBilinear2d(scale_factor=2 ** (i + j + 1))
                x = layer(x)
                y = layer(y)
                if j == 0: # skip first layer (no normalization)
                    continue
                diff = torch.abs(x - y)
                diff = torch.mean(diff, 1).unsqueeze(1)
                loss.append(up(diff))

            # content loss of channel discriminator
            for j, discs in enumerate(self.channel_discs.values()):
                up = nn.UpsamplingBilinear2d(scale_factor=2 ** (i + j + 1))
                x = input_real[:, j:j + 1]
                y = input_fake[:, j:j + 1]
                for k, layer in enumerate(discs[i][:-skip_last]):
                    x = layer(x)
                    y = layer(y)
                    if k == 0:  # skip first layer (no normalization)
                        continue
                    diff = torch.abs(x - y)
                    diff = torch.mean(diff, 1).unsqueeze(1)
                    loss.append(up(diff))

            input_real = self.downsample(input_real)
            input_fake = self.downsample(input_fake)
        return torch.mean(torch.cat(loss, 1), 1).unsqueeze(1)


########################## Encoder / Decoder ##########################

class NoiseEstimator(nn.Module):
    def __init__(self, input_dim, dim, noise_dim, depth, **kwargs):
        super().__init__()

        model = []
        model += [Conv2dBlock(input_dim, dim, 4, 2, 1, norm='none', activation=kwargs['activation'], pad_type=kwargs['pad_type'])]
        for i in range(depth - 1):
            model += [Conv2dBlock(dim, dim * 2, 4, 2, 1, **kwargs)]
            dim *= 2
        model += [nn.Conv2d(dim, noise_dim, 1, 1, 0), nn.Sigmoid()]
        self.model = nn.Sequential(*model)

    def forward(self, x):
        return self.model(x)

class DownBlock(nn.Module):

    def __init__(self, dim, out_dim, n_convs, **kwargs):
        super().__init__()
        assert n_convs >= 1, 'Invalid configuration, requires at least 1 convolution block'
        self.convs = nn.Sequential(*[Conv2dBlock(dim, dim, 3, 1, 1, **kwargs) for _ in range(n_convs)])
        self.down = Conv2dBlock(dim, out_dim, 3, 2, 1, **kwargs)
        self.module_list = nn.ModuleList([self.convs, self.down])

    def forward(self, x):
        x = self.convs(x)
        skip = x
        x = self.down(x)
        return x, skip


class UpBlock(nn.Module):

    def __init__(self, in_dim, out_dim, n_convs, skip_connection=True, **kwargs):
        super().__init__()
        assert n_convs >= 1, 'Invalid configuration, requires at least 1 convolution block'
        self.up = nn.Sequential(nn.UpsamplingBilinear2d(scale_factor=2), Conv2dBlock(in_dim, out_dim, 3, 1, 1, **kwargs))
        self.convs = nn.Sequential(
            *[Conv2dBlock(out_dim * 2 if i == 0 and skip_connection else out_dim, out_dim, 3, 1, 1, **kwargs)
              for i in range(n_convs)])
        self.module_list = nn.ModuleList([self.up, self.convs])

    def forward(self, x, skip=None):
        x = self.up(x)
        if skip is not None:
            x = torch.cat([x, skip], dim=1)
        x = self.convs(x)
        return x

class CoreBlock(nn.Module):

    def __init__(self, in_dim, out_dim, n_convs, **kwargs):
        super().__init__()
        assert n_convs > 1, 'Invalid configuration, requires at least 1 convolution block'
        module = [Conv2dBlock(in_dim, out_dim, 3, 1, 1, **kwargs)]
        module += [Conv2dBlock(out_dim, out_dim, 3, 1, 1, **kwargs) for _ in range(n_convs - 1)]
        self.module = nn.Sequential(*module)

    def forward(self, x):
        return self.module(x)


class Conv2dBlock(nn.Module):
    def __init__(self, input_dim, output_dim, kernel_size, stride,
                 padding=0, norm='none', activation='relu', pad_type='zero', transpose=False):
        super().__init__()
        self.use_bias = True
        # initialize padding
        if transpose:
            self.pad = nn.ZeroPad2d(0)
        elif pad_type == 'reflect':
            self.pad = nn.ReflectionPad2d(padding)
        elif pad_type == 'replicate':
            self.pad = nn.ReplicationPad2d(padding)
        elif pad_type == 'zero':
            self.pad = nn.ZeroPad2d(padding)
        else:
            assert 0, "Unsupported padding type: {}".format(pad_type)

        # initialize normalization
        norm_dim = output_dim
        if norm == 'bn':
            self.norm = nn.BatchNorm2d(norm_dim)
        elif norm == 'in':
            self.norm = nn.InstanceNorm2d(norm_dim)
        elif norm == 'in_rs':
            self.norm = nn.InstanceNorm2d(norm_dim, track_running_stats=True, momentum=0.01)
        elif norm == 'in_aff':
            self.norm = nn.InstanceNorm2d(norm_dim, track_running_stats=False, affine=True)
        elif norm == 'in_rs_aff':
            self.norm = nn.InstanceNorm2d(norm_dim, track_running_stats=True, momentum=0.01, affine=True)
        elif norm == 'ln':
            self.norm = LayerNorm(norm_dim)
        elif norm == 'none' or norm == 'sn':
            self.norm = None
        else:
            assert 0, "Unsupported normalization: {}".format(norm)

        # initialize activation
        if activation == 'relu':
            self.activation = nn.ReLU(inplace=False)
        elif activation == 'lrelu':
            self.activation = nn.LeakyReLU(0.2, inplace=False)
        elif activation == 'prelu':
            self.activation = nn.PReLU()
        elif activation == 'selu':
            self.activation = nn.SELU(inplace=False)
        elif activation == 'tanh':
            self.activation = nn.Tanh()
        elif activation == 'none':
            self.activation = None
        else:
            assert 0, "Unsupported activation: {}".format(activation)

        # initialize convolution
        conv = nn.Conv2d(input_dim, output_dim, kernel_size, stride, bias=self.use_bias) if not transpose \
            else nn.ConvTranspose2d(input_dim, output_dim, kernel_size, stride, padding=padding, bias=self.use_bias)
        if norm == 'sn':
            self.conv = SpectralNorm(conv)
        else:
            self.conv = conv

    def init_conv(self, conv):
        init.kaiming_normal(conv.weight)
        if conv.bias is not None:
            conv.bias.data.zero_()

    def forward(self, x):
        x = self.conv(self.pad(x))
        if self.norm:
            x = self.norm(x)
        if self.activation:
            x = self.activation(x)
        return x

class SpectralNorm(nn.Module):
    """
    Based on the paper "Spectral Normalization for Generative Adversarial Networks" by Takeru Miyato, Toshiki Kataoka, Masanori Koyama, Yuichi Yoshida
    and the Pytorch implementation https://github.com/christiancosgrove/pytorch-spectral-normalization-gan
    """

    def __init__(self, module, name='weight', power_iterations=1):
        super(SpectralNorm, self).__init__()
        self.module = module
        self.name = name
        self.power_iterations = power_iterations
        if not self._made_params():
            self._make_params()

    def _update_u_v(self):
        u = getattr(self.module, self.name + "_u")
        v = getattr(self.module, self.name + "_v")
        w = getattr(self.module, self.name + "_bar")

        height = w.data.shape[0]
        for _ in range(self.power_iterations):
            v.data = l2normalize(torch.mv(torch.t(w.view(height, -1).data), u.data))
            u.data = l2normalize(torch.mv(w.view(height, -1).data, v.data))

        # sigma = torch.dot(u.data, torch.mv(w.view(height,-1).data, v.data))
        sigma = u.dot(w.view(height, -1).mv(v))
        setattr(self.module, self.name, w / sigma.expand_as(w))

    def _made_params(self):
        try:
            u = getattr(self.module, self.name + "_u")
            v = getattr(self.module, self.name + "_v")
            w = getattr(self.module, self.name + "_bar")
            return True
        except AttributeError:
            return False

    def _make_params(self):
        w = getattr(self.module, self.name)

        height = w.data.shape[0]
        width = w.view(height, -1).data.shape[1]

        u = nn.Parameter(w.data.new(height).normal_(0, 1), requires_grad=False)
        v = nn.Parameter(w.data.new(width).normal_(0, 1), requires_grad=False)
        u.data = l2normalize(u.data)
        v.data = l2normalize(v.data)
        w_bar = nn.Parameter(w.data)

        del self.module._parameters[self.name]

        self.module.register_parameter(self.name + "_u", u)
        self.module.register_parameter(self.name + "_v", v)
        self.module.register_parameter(self.name + "_bar", w_bar)

    def forward(self, *args):
        self._update_u_v()
        return self.module.forward(*args)

class BatchStatistic(nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, x):
        return torch.cat([x,
                          torch.mean(x, [0, 2, 3], keepdim=True) * torch.ones_like(x),
                          torch.std(x, [0, 2, 3], keepdim=True) * torch.ones_like(x)], 1)

def l2normalize(v, eps=1e-12):
    return v / (v.norm() + eps)



import argparse
import logging
import os

import torch
from torch.utils.data import DataLoader

from itipy.callback import PlotBAB, PlotABA, VariationPlotBA, HistoryCallback, ProgressCallback, \
    SaveCallback
from itipy.data.dataset import KSOFlatDataset, StorageDataset, KSOFilmDataset
from itipy.data.editor import RandomPatchEditor
from itipy.trainer import Trainer, loop

parser = argparse.ArgumentParser(description='Train KSO Film-To-CCD translations')
parser.add_argument('--base_dir', type=str, help='path to the results directory.')

parser.add_argument('--kso_path', type=str, help='path to the high-quality KSO data.')
parser.add_argument('--film_path', type=str, help='path to the film KSO data.')
parser.add_argument('--kso_converted_path', type=str, help='path to store the converted KSO data.')
parser.add_argument('--film_converted_path', type=str, help='path to store the converted KSO-film data.')
parser.add_argument('--resolution', type=int, help='resolution of the images (default=512).', default=512,
                    required=False)

args = parser.parse_args()

base_dir = args.base_dir
resolution = args.resolution
kso_path = args.kso_path
film_path = args.film_path
kso_converted_path = args.kso_converted_path
film_converted_path = args.film_converted_path

prediction_dir = os.path.join(base_dir, 'prediction')
os.makedirs(prediction_dir, exist_ok=True)

logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler("{0}/{1}.log".format(base_dir, "info_log")),
        logging.StreamHandler()
    ])

# Init Model
trainer = Trainer(1, 1)
trainer.cuda()
trainer.train()
start_it = trainer.resume(base_dir)

# Init Dataset
ccd_dataset = KSOFlatDataset(kso_path, resolution, months=list(range(12)))
film_dataset = KSOFilmDataset(film_path, resolution, months=list(range(12)))
ccd_storage = StorageDataset(ccd_dataset, kso_converted_path, ext_editors=[RandomPatchEditor((256, 256))])
film_storage = StorageDataset(film_dataset, film_converted_path, ext_editors=[RandomPatchEditor((256, 256))])

ccd_plot = StorageDataset(ccd_dataset, kso_converted_path)
film_plot = StorageDataset(film_dataset, film_converted_path)

kso_ccd_iterator = loop(DataLoader(ccd_storage, batch_size=1, shuffle=True, num_workers=8))
kso_film_iterator = loop(DataLoader(film_storage, batch_size=1, shuffle=True, num_workers=8))

# Init Plot Callbacks
history = HistoryCallback(trainer, base_dir)
progress = ProgressCallback(trainer)
save = SaveCallback(trainer, base_dir)

plot_settings_A = {"cmap": "gray", "title": "Quality A", 'vmin': -1, 'vmax': 1}
plot_settings_B = {"cmap": "gray", "title": "Quality B", 'vmin': -1, 'vmax': 1}

log_iteration = 1000
bab_callback = PlotBAB(ccd_plot.sample(3), trainer, prediction_dir, log_iteration=log_iteration,
                       plot_settings_A=plot_settings_A, plot_settings_B=plot_settings_B)

aba_callback = PlotABA(film_plot.sample(3), trainer, prediction_dir, log_iteration=log_iteration,
                       plot_settings_A=plot_settings_A, plot_settings_B=plot_settings_B)

cutout_callback = PlotABA(film_storage.sample(6), trainer, prediction_dir, log_iteration=log_iteration,
                          plot_settings_A=plot_settings_A, plot_settings_B=plot_settings_B, plot_id='CUTOUT')

v_callback = VariationPlotBA(ccd_plot.sample(3), trainer, prediction_dir, 4, log_iteration=log_iteration,
                             plot_settings_A=plot_settings_A, plot_settings_B=plot_settings_B)

aba_callback.call(0)
bab_callback.call(0)
cutout_callback.call(0)
callbacks = [history, progress, save, bab_callback, aba_callback, cutout_callback, v_callback]

# Start training
for it in range(start_it, int(1e8)):
    if it > 100000:
        trainer.gen_ab.eval()  # fix running stats
        trainer.gen_ba.eval()  # fix running stats
    x_a, x_b = next(kso_film_iterator), next(kso_ccd_iterator)
    x_a, x_b = x_a.float().cuda().detach(), x_b.float().cuda().detach()
    #
    trainer.discriminator_update(x_a, x_b)
    trainer.generator_update(x_a, x_b)
    torch.cuda.synchronize()
    #
    for callback in callbacks:
        callback(it)



import argparse
import glob
import os
from multiprocessing import Pool
from warnings import simplefilter

import numpy as np
import pandas as pd
from sunpy.map import Map, all_coordinates_from_map
from tqdm import tqdm


def classify(file):
    simplefilter('ignore')
    s_map = Map(file)
    hpc_coords = all_coordinates_from_map(s_map)
    r = np.sqrt(hpc_coords.Tx ** 2 + hpc_coords.Ty ** 2) / s_map.rsun_obs
    if np.any(r > 1):
        return (file, 'limb', s_map.date)
    if np.std(s_map.data / s_map.meta['EXPTIME']) < 2500:
        return (file, 'quiet', s_map.date)
    return (file, 'feature', s_map.date)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Coarse classification of Hinode files for optimized data sampling')
    parser.add_argument('--hinode_dir', type=str, help='path to the Hinode data.')
    parser.add_argument('--n_workers', type=str, help='number of parallel threads.', required=False, default=4)
    parser.add_argument('--result_path', type=str, help='path to print the result csv.')
    args = parser.parse_args()

    hinode_files = sorted(glob.glob(os.path.join(args.hinode_dir, '*.fits')))
    with Pool(args.n_workers) as p:
        file_list = [d for d in tqdm(p.imap_unordered(classify, hinode_files), total=len(hinode_files))]

    pd.DataFrame(data=file_list, columns=['file', 'classification', 'date']).to_csv(args.result_path, index=False)



import argparse
import logging
import multiprocessing
import os
from datetime import datetime, timedelta
from urllib import request

import drms
import numpy as np
import pandas as pd
from astropy.io import fits
from sunpy.io.fits import header_to_fits
from sunpy.util import MetaDict


class HMIContinuumDownloader:

    def __init__(self, ds_path, email, num_worker_threads=4, ignore_quality=False, series='hmi.Ic_720s'):
        self.series = series
        self.ignore_quality = ignore_quality
        self.ds_path = ds_path
        self.n_workers = num_worker_threads
        os.makedirs(ds_path, exist_ok=True)

        logging.basicConfig(
            level=logging.INFO,
            handlers=[
                logging.FileHandler("{0}/{1}.log".format(ds_path, "info_log")),
                logging.StreamHandler()
            ])

        self.drms_client = drms.Client(email=email, verbose=False)

    def download(self, data):
        header, segment, t = data
        map_path = os.path.join(self.ds_path, '%s.fits' % t.isoformat('T', timespec='seconds'))
        if os.path.exists(map_path):
            return map_path
        # load map
        url = 'http://jsoc.stanford.edu' + segment
        request.urlretrieve(url, filename=map_path)

        header['DATE_OBS'] = header['DATE__OBS']
        header = header_to_fits(MetaDict(header))
        with fits.open(map_path, 'update') as f:
            hdr = f[1].header
            for k, v in header.items():
                if pd.isna(v):
                    continue
                hdr[k] = v
            f.verify('silentfix')

        return map_path

    def fetchDates(self, dates):
        header_info = []
        logging.info('Fetch header information')
        for date in dates:
            try:
                header_info += self.fetchData(date)
            except Exception as ex:
                print(ex)
                logging.error('Unable to download: %s' % date.isoformat())

        logging.info('Download data')
        with multiprocessing.Pool(self.n_workers) as p:
            files = p.map(self.download, header_info)
        return files

    def fetchData(self, time):
        # query
        time_param = '%sZ' % time.isoformat('_', timespec='seconds')
        ds_hmi = '%s[%s]{continuum}' % (self.series, time_param)
        keys_hmi = self.drms_client.keys(ds_hmi)
        header_hmi, segment_hmi = self.drms_client.query(ds_hmi, key=','.join(keys_hmi), seg='continuum')
        if len(header_hmi) != 1 or (np.any(header_hmi.QUALITY != 0) and not self.ignore_quality):
            raise Exception('No valid data found!')

        data = [(h.to_dict(), s, time) for (idx, h), s in zip(header_hmi.iterrows(), segment_hmi.continuum)]
        return data


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Download SDO/HMI data')
    parser.add_argument('--download_dir', type=str, help='path to the download directory.')
    parser.add_argument('--email', type=str, help='registered email address for JSOC.')
    args = parser.parse_args()

    fetcher = HMIContinuumDownloader(ds_path=args.download_dir, email=args.email)
    for y in range(2010, 2023):
        dates = [datetime(y, 1, 1) + i * timedelta(days=1) for i in
                 range((datetime(y + 1, 1, 1) - datetime(y, 1, 1)) // timedelta(days=1))]
        dates = [d for d in dates if d.month in [2, 3, 4, 5, 6, 7, 8, 9, 11, 12]]
        fetcher.fetchDates(dates)



import argparse
import logging
import multiprocessing
import os
from datetime import timedelta, datetime
from urllib import request

import drms
import numpy as np
import pandas as pd
from astropy.io import fits
from sunpy.io.fits import header_to_fits
from sunpy.util import MetaDict


class SDODownloader:

    def __init__(self, base_path, email, wavelengths=['171', '193', '211', '304'], n_workers=5):
        self.ds_path = base_path
        self.wavelengths = [str(wl) for wl in wavelengths]
        self.n_workers = n_workers
        [os.makedirs(os.path.join(base_path, wl), exist_ok=True) for wl in self.wavelengths + ['6173']]

        self.drms_client = drms.Client(email=email, verbose=False)

    def download(self, sample):
        header, segment, t = sample
        try:
            dir = os.path.join(self.ds_path, '%d' % header['WAVELNTH'])
            map_path = os.path.join(dir, '%s.fits' % t.isoformat('T', timespec='seconds'))
            if os.path.exists(map_path):
                return map_path
            # load map
            url = 'http://jsoc.stanford.edu' + segment
            request.urlretrieve(url, filename=map_path)

            header['DATE_OBS'] = header['DATE__OBS']
            header = header_to_fits(MetaDict(header))
            with fits.open(map_path, 'update') as f:
                hdr = f[1].header
                for k, v in header.items():
                    if pd.isna(v):
                        continue
                    hdr[k] = v
                f.verify('silentfix')

            return map_path
        except Exception as ex:
            logging.info('Download failed: %s (requeue)' % header['DATE__OBS'])
            logging.info(ex)
            raise ex

    def downloadDate(self, date):
        id = date.isoformat()

        logging.info('Start download: %s' % id)
        # query Magnetogram
        time_param = '%sZ' % date.isoformat('_', timespec='seconds')
        ds_hmi = 'hmi.M_720s[%s]{magnetogram}' % time_param
        keys_hmi = self.drms_client.keys(ds_hmi)
        header_hmi, segment_hmi = self.drms_client.query(ds_hmi, key=','.join(keys_hmi), seg='magnetogram')
        if len(header_hmi) != 1 or np.any(header_hmi.QUALITY != 0):
            self.fetchDataFallback(date)
            return

        # query EUV
        time_param = '%sZ' % date.isoformat('_', timespec='seconds')
        ds_euv = 'aia.lev1_euv_12s[%s][%s]{image}' % (time_param, ','.join(self.wavelengths))
        keys_euv = self.drms_client.keys(ds_euv)
        header_euv, segment_euv = self.drms_client.query(ds_euv, key=','.join(keys_euv), seg='image')
        if len(header_euv) != len(self.wavelengths) or np.any(header_euv.QUALITY != 0):
            self.fetchDataFallback(date)
            return

        queue = []
        for (idx, h), s in zip(header_hmi.iterrows(), segment_hmi.magnetogram):
            queue += [(h.to_dict(), s, date)]
        for (idx, h), s in zip(header_euv.iterrows(), segment_euv.image):
            queue += [(h.to_dict(), s, date)]

        with multiprocessing.Pool(self.n_workers) as p:
            p.map(self.download, queue)
        logging.info('Finished: %s' % id)

    def fetchDataFallback(self, date):
        id = date.isoformat()

        logging.info('Fallback download: %s' % id)
        # query Magnetogram
        t = date - timedelta(hours=6)
        ds_hmi = 'hmi.M_720s[%sZ/12h@720s]{magnetogram}' % t.replace(tzinfo=None).isoformat('_', timespec='seconds')
        keys_hmi = self.drms_client.keys(ds_hmi)
        header_tmp, segment_tmp = self.drms_client.query(ds_hmi, key=','.join(keys_hmi), seg='magnetogram')
        assert len(header_tmp) != 0, 'No data found!'
        date_str = header_tmp['DATE__OBS'].replace('MISSING', '').str.replace('60', '59')  # fix date format
        date_diff = np.abs(pd.to_datetime(date_str).dt.tz_localize(None) - date)
        # sort and filter
        header_tmp['date_diff'] = date_diff
        header_tmp.sort_values('date_diff')
        segment_tmp['date_diff'] = date_diff
        segment_tmp.sort_values('date_diff')
        cond_tmp = header_tmp.QUALITY == 0
        header_tmp = header_tmp[cond_tmp]
        segment_tmp = segment_tmp[cond_tmp]
        assert len(header_tmp) > 0, 'No valid quality flag found'
        # replace invalid
        header_hmi = header_tmp.iloc[0].drop('date_diff')
        segment_hmi = segment_tmp.iloc[0].drop('date_diff')
        ############################################################
        # query EUV
        header_euv, segment_euv = [], []
        t = date - timedelta(hours=6)
        for wl in self.wavelengths:
            euv_ds = 'aia.lev1_euv_12s[%sZ/12h@12s][%s]{image}' % (
                t.replace(tzinfo=None).isoformat('_', timespec='seconds'), wl)
            keys_euv = self.drms_client.keys(euv_ds)
            header_tmp, segment_tmp = self.drms_client.query(euv_ds, key=','.join(keys_euv), seg='image')
            assert len(header_tmp) != 0, 'No data found!'
            date_str = header_tmp['DATE__OBS'].replace('MISSING', '').str.replace('60', '59')  # fix date format
            date_diff = (pd.to_datetime(date_str).dt.tz_localize(None) - date).abs()
            # sort and filter
            header_tmp['date_diff'] = date_diff
            header_tmp.sort_values('date_diff')
            segment_tmp['date_diff'] = date_diff
            segment_tmp.sort_values('date_diff')
            cond_tmp = header_tmp.QUALITY == 0
            header_tmp = header_tmp[cond_tmp]
            segment_tmp = segment_tmp[cond_tmp]
            assert len(header_tmp) > 0, 'No valid quality flag found'
            # replace invalid
            header_euv.append(header_tmp.iloc[0].drop('date_diff'))
            segment_euv.append(segment_tmp.iloc[0].drop('date_diff'))

        queue = []
        queue += [(header_hmi.to_dict(), segment_hmi.magnetogram, date)]
        for h, s in zip(header_euv, segment_euv):
            queue += [(h.to_dict(), s.image, date)]

        with multiprocessing.Pool(self.n_workers) as p:
            p.map(self.download, queue)

        logging.info('Finished: %s' % id)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Download SDO data from JSOC with quality check and fallback')
    parser.add_argument('--download_dir', type=str, help='path to the download directory.')
    parser.add_argument('--email', type=str, help='registered email address for JSOC.')

    args = parser.parse_args()
    download_dir = args.download_dir

    [os.makedirs(os.path.join(download_dir, str(c)), exist_ok=True) for c in [171, 193, 211, 304, 6173]]
    downloader = SDODownloader(base_path=download_dir, email=args.email)
    start_date = datetime(2010, 5, 13)
    for d in [start_date + i * timedelta(hours=6) for i in
              range((datetime.now() - start_date) // timedelta(hours=6))]:
        downloader.downloadDate(d)



import argparse
import logging
import os

from sunpy.visualization.colormaps import cm

from itipy.data.dataset import SDODataset, SOHODataset, StorageDataset
from itipy.data.editor import RandomPatchEditor, LambdaEditor
from itipy.train.model import DiscriminatorMode
from itipy.trainer import Trainer

parser = argparse.ArgumentParser(description='Train SOHO-To-SDO translations without magnetograms')
parser.add_argument('--base_dir', type=str, help='path to the results directory.')

parser.add_argument('--sdo_path', type=str, help='path to the SDO data.')
parser.add_argument('--soho_path', type=str, help='path to the SOHO data.')
parser.add_argument('--sdo_converted_path', type=str, help='path to store the converted SDO data.')
parser.add_argument('--soho_converted_path', type=str, help='path to store the converted SOHO data.')

args = parser.parse_args()
base_dir = args.base_dir

sdo_path = args.sdo_path
soho_path = args.soho_path
sdo_converted_path = args.sdo_converted_path
soho_converted_path = args.soho_converted_path

prediction_dir = os.path.join(base_dir, 'prediction')
os.makedirs(prediction_dir, exist_ok=True)

log_iteration = 1000
logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler("{0}/{1}.log".format(base_dir, "info_log")),
        logging.StreamHandler()
    ])

# Init Model
trainer = Trainer(4, 4, upsampling=1, discriminator_mode=DiscriminatorMode.CHANNELS,
                  lambda_diversity=0, norm='in_rs_aff')
trainer.cuda()

channel_editor = LambdaEditor(lambda x, **kwargs: x[:-1])

# Init Dataset
train_months = list(range(2, 10))
test_months = [11, 12]

sdo_train = SDODataset(sdo_path, patch_shape=(1024, 1024), resolution=2048, months=train_months,
                       years=list(range(2011, 2021)))
sdo_train = StorageDataset(sdo_train, sdo_converted_path,
                           ext_editors=[channel_editor, RandomPatchEditor((256, 256))])

soho_train = SOHODataset(soho_path, resolution=1024, months=train_months)

soho_train = StorageDataset(soho_train, soho_converted_path,
                            ext_editors=[channel_editor, RandomPatchEditor((128, 128))])

sdo_valid = SDODataset(sdo_path, patch_shape=(1024, 1024), resolution=2048, months=test_months, limit=100)
sdo_valid = StorageDataset(sdo_valid, sdo_converted_path, ext_editors=[channel_editor, RandomPatchEditor((256, 256))])

soho_valid = SOHODataset(soho_path, resolution=1024, months=test_months, limit=100)
soho_valid = StorageDataset(soho_valid, soho_converted_path,
                            ext_editors=[channel_editor, RandomPatchEditor((512, 512))])

plot_settings_A = [
    {"cmap": cm.sdoaia171, "title": "EIT 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "EIT 195", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "EIT 284", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "EIT 304", 'vmin': -1, 'vmax': 1},
]
plot_settings_B = [
    {"cmap": cm.sdoaia171, "title": "AIA 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "AIA 193", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "AIA 211", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "AIA 304", 'vmin': -1, 'vmax': 1},
]

trainer.startBasicTraining(base_dir, soho_train, sdo_train, soho_valid, sdo_valid, plot_settings_A, plot_settings_B)



import argparse
from datetime import timedelta, datetime

import numpy as np
from dateutil.parser import parse


parser = argparse.ArgumentParser(description='Prepare Hinode download script')
parser.add_argument('--sh_path', type=str, help='the path to the sh file including all available observations.')
parser.add_argument('--filtered_path', type=str, help='path to the result sh file.')

args = parser.parse_args()

file_name = args.sh_path
filtered_file_name = args.filtered_path

with open(file_name) as f:
    lines = np.array(f.readlines())

lines = np.array([l for l in lines if 'FGFOCUS' not in l])
new_lines = lines[:2]

dates = [parse(l[-23:-8].replace('_', 'T')) for l in lines[2:]]
date_diff = np.diff(dates, append=datetime.now())

new_lines = np.concatenate([new_lines, lines[2:][date_diff > timedelta(minutes=30)]])

with open(filtered_file_name, 'w') as f:
    f.writelines(new_lines)



import argparse
import glob
import os
from multiprocessing import Pool

import numpy as np
import pandas as pd
from dateutil.parser import parse
from tqdm import tqdm

from itipy.data.editor import LoadMapEditor, NormalizeRadiusEditor, MapToDataEditor, EITCheckEditor, RemoveOffLimbEditor, \
    AIAPrepEditor, SECCHIPrepEditor

parser = argparse.ArgumentParser(description='Estimate the mean and std of SOHO, STEREO and SDO for calibration.')
parser.add_argument('--soho_path', type=str, help='the path to the soho files.')
parser.add_argument('--stereo_path', type=str, help='the path to the stereo files.')
parser.add_argument('--sdo_path', type=str, help='the path to the sdo files.')
parser.add_argument('--evaluation_path', type=str, help='the path for printing the results.')

args = parser.parse_args()

soho_path = args.soho_path
sdo_path = args.sdo_path
stereo_path = args.stereo_path
evaluation_path = args.evaluation_path

soho_channels = ['171', '195', '284', '304']
sdo_channels = ['171', '193', '211', '304']

eit_files = [sorted(glob.glob(os.path.join(soho_path, c, '*.fits'))) for c in soho_channels]
secchi_files = [sorted(glob.glob(os.path.join(stereo_path, c, '*.fits'))) for c in soho_channels]
aia_files = [sorted(glob.glob(os.path.join(sdo_path, c, '*.fits'))) for c in sdo_channels]


def getEITData(f):
    s_map, _ = LoadMapEditor().call(f)
    s_map = EITCheckEditor().call(s_map)
    s_map = NormalizeRadiusEditor(1024).call(s_map)
    s_map = RemoveOffLimbEditor(fill_value=np.nan).call(s_map)
    data, _ = MapToDataEditor().call(s_map)
    return data


def getAIAData(f):
    s_map, _ = LoadMapEditor().call(f)
    s_map = NormalizeRadiusEditor(1024).call(s_map)
    s_map = AIAPrepEditor(calibration='auto').call(s_map)
    s_map = RemoveOffLimbEditor(fill_value=np.nan).call(s_map)
    data, _ = MapToDataEditor().call(s_map)
    return data


def getSECCHIData(f):
    s_map, _ = LoadMapEditor().call(f)
    deg = [-9.42497209e-05, 2.27153104e+00] if s_map.wavelength.value == 304 else None
    s_map = SECCHIPrepEditor(degradation=deg).call(s_map)
    s_map = NormalizeRadiusEditor(1024).call(s_map)
    s_map = RemoveOffLimbEditor(fill_value=np.nan).call(s_map)
    data, _ = MapToDataEditor().call(s_map)
    return data


def filter_files(files, years=None):
    dates = [parse(os.path.basename(f).split('.')[0]) for f in files]
    files, dates = zip(*[(f, d) for f, d in zip(files, dates) if d.month in list(range(2, 10))])
    if years:
        files, dates = zip(*[(f, d) for f, d in zip(files, dates) if d.year in years])
    df = pd.DataFrame({'date': dates, 'file': files})
    df = df.set_index('date').groupby(pd.Grouper(freq='10D')).first()
    return df[~pd.isna(df['file'])]['file'].tolist()


eit_hist = {}
for c, c_files in zip(soho_channels, eit_files):
    c_files = filter_files(c_files, years=list(range(1996, 2010)))
    with Pool(12) as p:
        data = [np.ravel(m) for m in tqdm(p.imap_unordered(getEITData, c_files), total=len(c_files))]
        data = np.concatenate(data)
    threshold = np.nanmedian(data) + np.nanstd(data)
    data[data > threshold] = np.nan
    eit_hist[c] = [np.nanmean(data), np.nanstd(data)]

secchi_hist = {}
for c, c_files in zip(soho_channels, secchi_files):
    c_files = c_files[::len(c_files) // 100]
    with Pool(12) as p:
        data = [np.ravel(m) for m in tqdm(p.imap_unordered(getSECCHIData, c_files), total=len(c_files))]
        data = np.concatenate(data)
    threshold = np.nanmedian(data) + np.nanstd(data)
    data[data > threshold] = np.nan
    secchi_hist[c] = [np.nanmean(data), np.nanstd(data)]

aia_hist = {}
for c, c_files in zip(sdo_channels, aia_files):
    c_files = c_files[::len(c_files) // 100]
    with Pool(12) as p:
        data = [np.ravel(m) for m in tqdm(p.imap_unordered(getAIAData, c_files), total=len(c_files))]
        data = np.concatenate(data)
    threshold = np.nanmedian(data) + np.nanstd(data)
    data[data > threshold] = np.nan
    aia_hist[c] = [np.nanmean(data), np.nanstd(data)]

with open(os.path.join(evaluation_path, 'calibration.txt'), 'w') as f:
    print('EIT', file=f)
    print(eit_hist, file=f)
    print('SECCHI', file=f)
    print(secchi_hist, file=f)
    print('AIA', file=f)
    print(aia_hist, file=f)



import argparse
import logging
import os
import shutil
from datetime import timedelta, datetime
from multiprocessing import Pool
from urllib.request import urlopen
from warnings import simplefilter

import drms
import numpy as np
import pandas as pd
from astropy import units as u
from astropy.io.fits import getheader, HDUList
from dateutil.relativedelta import relativedelta
from sunpy.map import Map
from sunpy.net import Fido, attrs as a
from tqdm import tqdm


class SOHODownloader:

    def __init__(self, base_path):
        self.base_path = base_path
        self.wavelengths = [171, 195, 284, 304]
        self.dirs = ['mag', '171', '195', '284', '304', ]
        [os.makedirs(os.path.join(base_path, dir), exist_ok=True) for dir in self.dirs]

    def downloadDate(self, date):
        files = []
        try:
            # Download EIT
            for wl in self.wavelengths:
                files += [self.downloadEIT(date, wl)]
            # Download MDI
            if date < datetime(2011, 4, 12):  # No MDI data past 2011
                files += [self.downloadMDI(date)]
            logging.info('Download complete %s' % date.isoformat())
        except Exception as ex:
            logging.error('Unable to download %s: %s' % (date.isoformat(), str(ex)))
            [os.remove(f) for f in files]

    def downloadEIT(self, query_date, wl):
        file_path = os.path.join(self.base_path, str(wl), "%s.fits" % query_date.isoformat("T", timespec='seconds'))
        if os.path.exists(file_path):
            return file_path  # skip existing downloads (e.g. retry)
        #
        search = Fido.search(a.Time(query_date - timedelta(minutes=30), query_date + timedelta(minutes=30)),
                             a.Instrument('EIT'), a.Wavelength(wl * u.AA), a.Provider("SDAC"))
        assert search.file_num > 0, "No data found for %s (%s)" % (query_date.isoformat(), wl)
        search = sorted(search['vso'], key=lambda x: abs(x['Start Time'].datetime - query_date).total_seconds())
        for entry in search:
            files = Fido.fetch(entry, path=self.base_path, progress=False)
            if len(files) != 1:
                continue
            file = files[0]
            header = getheader(file)
            if header['NAXIS1'] != 1024 or header['NAXIS2'] != 1024 or 'N_MISSING_BLOCKS =    0' not in \
                    header['COMMENT'][-1]:
                os.remove(file)
                continue
            shutil.move(file, file_path)
            return file_path

        raise Exception("No valid file found for %s (%s)!" % (query_date.isoformat(), wl))

    def downloadMDI(self, download_date):
        simplefilter('ignore')
        file_path = os.path.join(self.base_path, self.dirs[0],
                                 "%s.fits" % download_date.isoformat("T", timespec='seconds'))
        if os.path.exists(file_path):
            return file_path
        time_param = '%sZ/192m' % (download_date - timedelta(minutes=96)).isoformat('_', timespec='seconds')
        ds_mdi = 'mdi.fd_M_96m_lev182[%s]{data}' % time_param
        keys_mdi = drms_client.keys(ds_mdi)
        headers = drms_client.query(ds_mdi, key=','.join(keys_mdi))
        headers['url'] = drms_client.export(ds_mdi).urls.url
        headers = headers[(headers['DATE__OBS'] != 'MISSING')]
        # ((headers['QUALITY'] == 0) | (headers['QUALITY'] == 512)) &
        headers['time_diff'] = (pd.to_datetime(headers['DATE__OBS']).dt.tz_localize(None) - download_date).abs()
        headers = headers.sort_values('time_diff')
        for i, header in headers.iterrows():
            url_request = urlopen(header.url)
            fits_data = url_request.read()
            hdul = HDUList.fromstring(fits_data)
            hdul.verify('silentfix')
            data = hdul[1].data
            header_dict = header.to_dict()
            header_dict['DATE_OBS'] = header_dict['DATE__OBS']
            del header_dict['time_diff']
            del header_dict['url']
            s_map = Map(data, header_dict)
            s_map.save(file_path)
            return file_path
        raise Exception('No valid magnetogram found!')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Download SOHO data')
    parser.add_argument('--download_dir', type=str, help='path to the download directory.')
    parser.add_argument('--n_workers', type=str, help='number of parallel threads.', required=False, default=4)

    args = parser.parse_args()
    base_path = args.download_dir
    n_workers = args.n_workers

    drms_client = drms.Client(email='robert.jarolim@uni-graz.at', verbose=False)
    download_util = SOHODownloader(base_path)
    logging.basicConfig(
        level=logging.INFO,
        handlers=[
            logging.StreamHandler()
        ])
    start_date = datetime(1996, 1, 1, 0, 0)
    end_date = datetime.now()
    num_months = (end_date.year - start_date.year) * 12 + (end_date.month - start_date.month)
    month_dates = [start_date + i * relativedelta(months=1) for i in range(num_months)]
    for date in month_dates:
            search = Fido.search(a.Time(date, date + relativedelta(months=1)),
                                 a.Provider("SDAC"), a.Instrument('EIT'),
                                 a.Wavelength(304 * u.AA))
            if search.file_num == 0:
                continue
            dates = search['vso']['Start Time']
            logging.info("TOTAL DATES (%s): %d" % (date.isoformat(), len(dates)))
            step = int(np.floor(len(dates) / 60)) if len(dates) > 60 else 1
            dates = [d.datetime for d in dates[::step]]

            with Pool(n_workers) as p:
                [None for _ in tqdm(p.imap_unordered(download_util.downloadDate, dates), total=len(dates))]



import argparse
import os
import sys
from ftplib import FTP

import pandas as pd


def searchHalpha(base_dir=""):
    print("SEARCHING:", base_dir)
    files = ftp.nlst(base_dir)
    for file in files:
        if os.path.splitext(file)[1] != "":
            if file.endswith(".fts.gz") and (valid_files is not None and
                                             (valid_files.file_name == os.path.basename(file)).any()):
                downloadFile(file)
            continue
        if file == base_dir:
            continue
        searchHalpha(file)


def downloadFile(filename):
    target_file_path = os.path.join(local_path, os.path.basename(filename))
    if os.path.exists(target_file_path):
        return
    print("DOWNLOADING", filename)
    file = open(target_file_path, 'wb')
    ftp.retrbinary('RETR ' + filename, file.write)
    file.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Download KSO Halpha data')
    parser.add_argument('--download_dir', type=str, help='path to the download directory.')
    parser.add_argument('--download_files_csv', type=str, help='csv file with basenames of files to download.',
                        required=False, default=None)
    args = parser.parse_args()

    local_path = args.download_dir
    valid_files = pd.read_csv(args.download_files_csv, index_col=0) if args.download_files_csv is not None else None

    os.makedirs(local_path, exist_ok=True)

    ftp = FTP('ftp.kso.ac.at')
    ftp.login('download', '9521treffen')
    searchHalpha("halpha4M/FITS/synoptic")
    ftp.quit()
